from collections import defaultdict

from robot.api.parsing import ModelTransformer, Comment, Token, EmptyLine, LibraryImport
from robot.libraries import STDLIBS

from robotidy.exceptions import InvalidParameterValueError


class OrderSettingsSection(ModelTransformer):
    """
    Order settings inside *** Settings *** section.

    Settings are grouped in following groups:
      - documentation (Documentation, Metadata),
      - imports (Library, Resource, Variables),
      - settings (Suite Setup and Teardown, Test Setup and Teardown, Test Timeout, Test Template),
      - tags (Force Tags, Default Tags)

    Then ordered by groups (according to ``group_order = documentation,imports,settings,tags`` order). Every
    group is separated by ``new_lines_between_groups = 1`` new lines.
    Settings are grouped inside group. Default order can be modified through following parameters:
      - ``documentation_order = documentation,metadata``
      - ``imports_order = preserved``
      - ``settings_order = suite_setup,suite_teardown,test_setup,test_teardown,test_timeout,test_template``

    By default order of imports is preserved. You can overwrite this behaviour:

        robotidy --configure OrderSettingsSections:imports_order=library,resource,variables

    You can also preserve order inside any group by passing ``preserved`` instead of setting names:

        robotidy --configure OrderSettingsSections:settings_order=preserved

    Setting names omitted from custom order will be removed from the file. In following example we are missing metadata
    therefore all metadata will be removed:

        robotidy --configure OrderSettingsSection:documentation_order=documentation

    Libraries are grouped into built in libraries and custom libraries.
    Parsing errors (such as Resources instead of Resource, duplicated settings) are moved to the end of section.

    See https://robotidy.readthedocs.io/en/latest/transformers/OrderSettingsSection.html for more examples.
    """

    def __init__(
        self,
        new_lines_between_groups: int = 1,
        group_order: str = None,
        documentation_order: str = None,
        imports_order: str = "preserved",
        settings_order: str = None,
        tags_order: str = None,
    ):
        self.last_section = None
        self.disabled_group = set()
        self.new_lines_between_groups = new_lines_between_groups
        self.group_order = self.parse_group_order(group_order)
        self.documentation_order = self.parse_order_in_group(
            "documentation",
            documentation_order,
            (Token.DOCUMENTATION, Token.METADATA),
            {"documentation": Token.DOCUMENTATION, "metadata": Token.METADATA},
        )
        self.imports_order = self.parse_order_in_group(
            "imports",
            imports_order,
            (Token.LIBRARY, Token.RESOURCE, Token.VARIABLES),
            {
                "library": Token.LIBRARY,
                "resource": Token.RESOURCE,
                "variables": Token.VARIABLES,
            },
        )
        self.settings_order = self.parse_order_in_group(
            "settings",
            settings_order,
            (
                Token.SUITE_SETUP,
                Token.SUITE_TEARDOWN,
                Token.TEST_SETUP,
                Token.TEST_TEARDOWN,
                Token.TEST_TIMEOUT,
                Token.TEST_TEMPLATE,
            ),
            {
                "suite_setup": Token.SUITE_SETUP,
                "suite_teardown": Token.SUITE_TEARDOWN,
                "test_setup": Token.TEST_SETUP,
                "test_teardown": Token.TEST_TEARDOWN,
                "test_timeout": Token.TEST_TIMEOUT,
                "test_template": Token.TEST_TEMPLATE,
            },
        )
        self.tags_order = self.parse_order_in_group(
            "tags",
            tags_order,
            (Token.FORCE_TAGS, Token.DEFAULT_TAGS),
            {"force_tags": Token.FORCE_TAGS, "default_tags": Token.DEFAULT_TAGS},
        )

    def parse_group_order(self, order):
        default = ("documentation", "imports", "settings", "tags")
        if order is None:
            return default
        if not order:
            return []
        parts = order.lower().split(",")
        if any(part not in default for part in parts):
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "group_order",
                order,
                f"Custom order should be provided in comma separated list with valid group names:\n{default}",
            )
        return parts

    def parse_order_in_group(self, name, order, default, mapping):
        if order is None:
            return default
        if not order:
            return []
        if order == "preserved":
            self.disabled_group.add(name)
            return default
        parts = order.lower().split(",")
        try:
            return [mapping[part] for part in parts]
        except KeyError:
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "order",
                order,
                f"Custom order should be provided in comma separated list with valid group names:\n{sorted(mapping.keys())}",
            )

    def visit_File(self, node):  # noqa
        self.last_section = node.sections[-1] if node.sections else None
        return self.generic_visit(node)

    def visit_SettingSection(self, node):  # noqa
        if not node.body:
            return
        if node is self.last_section and not isinstance(node.body[-1], EmptyLine):
            node.body[-1] = self.fix_eol(node.body[-1])
        comments, errors = [], []
        groups = defaultdict(list)
        for child in node.body:
            child_type = getattr(child, "type", None)
            if isinstance(child, Comment):
                comments.append(child)
            elif child_type in self.documentation_order:
                groups["documentation"].append((comments, child))
                comments = []
            elif child_type in self.imports_order:
                groups["imports"].append((comments, child))
                comments = []
            elif child_type in self.settings_order:
                groups["settings"].append((comments, child))
                comments = []
            elif child_type in self.tags_order:
                groups["tags"].append((comments, child))
                comments = []
            elif not isinstance(child, EmptyLine):
                errors.append(child)

        group_map = {
            "documentation": self.documentation_order,
            "imports": self.imports_order,
            "settings": self.settings_order,
            "tags": self.tags_order,
        }

        new_body = []
        empty_line = EmptyLine.from_params(eol="\n")
        order_of_groups = [group for group in self.group_order if group in groups]
        last_index = len(order_of_groups) - 1
        for index, group in enumerate(order_of_groups):
            unordered = groups[group]
            if group in self.disabled_group:
                for comment_lines, child in unordered:
                    new_body.extend(comment_lines)
                    new_body.append(child)
            else:
                if group == "imports":
                    unordered = self.sort_builtin_libs(unordered)
                order = group_map[group]
                for token_type in order:
                    for comment_lines, child in unordered:
                        if child.type == token_type:
                            new_body.extend(comment_lines)
                            new_body.append(child)
            if index != last_index:
                new_body.extend([empty_line] * self.new_lines_between_groups)

        # not recognized headers, parsing errors like Resources instead of Resource
        if errors:
            new_body.extend([empty_line] * self.new_lines_between_groups)
            new_body.extend(errors)
        new_body.extend(comments)
        if node is not self.last_section:
            new_body.append(empty_line)
        node.body = new_body
        return node

    @staticmethod
    def fix_eol(node):
        if not getattr(node, "tokens", None):
            return node
        if getattr(node.tokens[-1], "type", None) != Token.EOL:
            return node
        node.tokens = list(node.tokens[:-1]) + [Token(Token.EOL, "\n")]
        return node

    @staticmethod
    def sort_builtin_libs(statements):
        before, after = [], []
        for comments, statement in statements:
            if (
                isinstance(statement, LibraryImport)
                and statement.name
                and statement.name != "Remote"
                and statement.name in STDLIBS
            ):
                before.append((comments, statement))
            else:
                after.append((comments, statement))
        before = sorted(before, key=lambda x: x[1].name)
        return before + after



import ast
import difflib
import importlib.util
import re
from collections import Counter, defaultdict
from importlib import import_module
from pathlib import Path
from typing import Pattern, List, Tuple, Dict, Optional

from robot.api import Token
from robot.parsing.model.statements import EmptyLine

try:
    from robot.api.parsing import Variable
except ImportError:
    from robot.parsing.model.statements import Variable

from robot.version import VERSION as RF_VERSION
from packaging import version

from robocop.version import __version__
from robocop.exceptions import InvalidExternalCheckerError

ROBOT_VERSION = version.parse(RF_VERSION)
if not hasattr(ROBOT_VERSION, "major"):
    ROBOT_VERSION.major = ROBOT_VERSION.release[0]


def modules_in_current_dir(path, module_name):
    """Yield modules inside `path` parent directory"""
    yield from modules_from_path(Path(path).parent, module_name)


def modules_from_paths(paths):
    for path in paths:
        path_object = Path(path)
        if path_object.exists():
            if path_object.is_dir():
                yield from modules_from_paths(
                    [file for file in path_object.iterdir() if "__pycache__" not in str(file)]
                )
            else:
                spec = importlib.util.spec_from_file_location(path_object.stem, path_object)
                mod = importlib.util.module_from_spec(spec)

                spec.loader.exec_module(mod)
                yield mod
        else:
            # if it's not physical path, try to import from installed modules
            try:
                parent_name, *lib_name = path.rsplit(".", 1)
                if lib_name:
                    parent = __import__(parent_name, fromlist=lib_name)
                    mod = getattr(parent, "".join(lib_name))
                else:
                    mod = __import__(parent_name, None)
                yield mod
            except ImportError:
                raise InvalidExternalCheckerError(path) from None


def modules_from_path(path, module_name=None, relative="."):
    """Traverse current directory and yield python files imported as module"""
    if path.is_file():
        yield import_module(relative + path.stem, module_name)
    elif path.is_dir():
        for file in path.iterdir():
            if "__pycache__" in str(file):
                continue
            if file.suffix == ".py" and file.stem != "__init__":
                yield from modules_from_path(file, module_name, relative)


def normalize_robot_name(name: str, remove_prefix: Optional[str] = None) -> str:
    name = name.replace(" ", "").replace("_", "").lower() if name else ""
    if remove_prefix:
        return name[name.startswith(remove_prefix) and len(remove_prefix) :]
    return name


def normalize_robot_var_name(name: str) -> str:
    return normalize_robot_name(name)[2:-1] if name else ""


def keyword_col(node) -> int:
    return token_col(node, Token.KEYWORD)


def token_col(node, *token_type) -> int:
    if ROBOT_VERSION.major == 3:
        for tok_type in token_type:
            token = node.get_token(tok_type)
            if token is not None:
                break
        else:
            return 1
    else:
        token = node.get_token(*token_type)

    if token is None:
        return 1
    return token.col_offset + 1


def issues_to_lsp_diagnostic(issues) -> List[Dict]:
    return [
        {
            "range": {
                "start": {
                    "line": max(0, issue.line - 1),
                    "character": max(0, issue.col - 1),
                },
                "end": {
                    "line": max(0, issue.end_line - 1),
                    "character": max(0, issue.end_col - 1),
                },
            },
            "severity": issue.severity.diag_severity(),
            "code": issue.rule_id,
            "source": "robocop",
            "message": issue.desc,
            "codeDescription": {"href": f"https://robocop.readthedocs.io/en/{__version__}/rules.html#{issue.name}"},
        }
        for issue in issues
    ]


class AssignmentTypeDetector(ast.NodeVisitor):
    """Visitor for counting number and type of assignments"""

    def __init__(self):
        self.keyword_sign_counter = Counter()
        self.keyword_most_common = None
        self.variables_sign_counter = Counter()
        self.variables_most_common = None

    def visit_File(self, node):  # noqa
        self.generic_visit(node)
        if len(self.keyword_sign_counter) >= 2:
            self.keyword_most_common = self.keyword_sign_counter.most_common(1)[0][0]
        if len(self.variables_sign_counter) >= 2:
            self.variables_most_common = self.variables_sign_counter.most_common(1)[0][0]

    def visit_KeywordCall(self, node):  # noqa
        if node.assign:  # if keyword returns any value
            sign = self.get_assignment_sign(node.assign[-1])
            self.keyword_sign_counter[sign] += 1

    def visit_VariableSection(self, node):  # noqa
        for child in node.body:
            if not isinstance(child, Variable):
                continue
            var_token = child.get_token(Token.VARIABLE)
            sign = self.get_assignment_sign(var_token.value)
            self.variables_sign_counter[sign] += 1
        return node

    @staticmethod
    def get_assignment_sign(token_value):
        return token_value[token_value.find("}") + 1 :]


def parse_assignment_sign_type(value: str) -> str:
    types = {
        "none": "",
        "equal_sign": "=",
        "space_and_equal_sign": " =",
        "autodetect": "autodetect",
    }
    if value not in types:
        raise ValueError(
            f"Expected one of ('none', 'equal_sign', 'space_and_equal_sign', 'autodetect') but got '{value}' instead"
        )
    return types[value]


class RecommendationFinder:
    def find_similar(self, name, candidates):
        norm_name = self.normalize(name)
        norm_cand = self.get_normalized_candidates(candidates)
        matches = []
        for norm in norm_name:
            matches += self.find(norm, norm_cand.keys())
        if not matches:
            return ""
        matches = self.get_original_candidates(matches, norm_cand)
        suggestion = "Did you mean:\n"
        suggestion += "\n".join(f"    {match}" for match in matches)
        return suggestion

    def find(self, name, candidates, max_matches=5):
        """Return a list of close matches to `name` from `candidates`."""
        if not name or not candidates:
            return []
        cutoff = self._calculate_cutoff(name)
        return difflib.get_close_matches(name, candidates, n=max_matches, cutoff=cutoff)

    @staticmethod
    def _calculate_cutoff(string, min_cutoff=0.5, max_cutoff=0.85, step=0.03):
        """The longer the string the bigger required cutoff."""
        cutoff = min_cutoff + len(string) * step
        return min(cutoff, max_cutoff)

    @staticmethod
    def normalize(name):
        """
        Return tuple where first element is string created from sorted words in name,
        and second element is name without `-` and `_`.
        """
        norm = re.split("[-_ ]+", name)
        return " ".join(sorted(norm)), name.replace("-", "").replace("_", "")

    @staticmethod
    def get_original_candidates(candidates, norm_candidates):
        """Map found normalized candidates to unique original candidates."""
        return sorted(list(set(c for cand in candidates for c in norm_candidates[cand])))

    def get_normalized_candidates(self, candidates):
        """
        Thanks for normalizing and sorting we can find cases like this-is, thisis, this-is1 instead of is-this.
        Normalized names form dictionary that point to original names - we're using list because several names can
        have one common normalized name.
        Different normalization methods try to imitate possible mistakes done when typing name - different order,
        missing `-` etc.
        """
        norm = defaultdict(list)
        for cand in candidates:
            for norm_cand in self.normalize(cand):
                norm[norm_cand].append(cand)
        return norm


class TestTemplateFinder(ast.NodeVisitor):
    def __init__(self):
        self.templated = False

    def visit_TestTemplate(self, node):  # noqa
        self.templated = bool(node.value)


def is_suite_templated(model):
    finder = TestTemplateFinder()
    finder.visit(model)
    return finder.templated


def last_non_empty_line(node):
    for child in node.body[::-1]:
        if not isinstance(child, EmptyLine):
            return child.lineno
    return node.lineno


def next_char_is(string: str, i: int, char: str) -> bool:
    if not i < len(string) - 1:
        return False
    return string[i + 1] == char


def remove_robot_vars(name: str) -> str:
    var_start = set("$@%&")
    brackets = 0
    open_bracket, close_bracket = "", ""
    replaced = ""
    index = 0
    while index < len(name):
        if brackets:
            if name[index] == open_bracket:
                brackets += 1
            elif name[index] == close_bracket:
                brackets -= 1
            # check if next chars are not ['key']
            if not brackets and next_char_is(name, index, "["):
                brackets += 1
                index += 1
                open_bracket, close_bracket = "[", "]"
        # it looks for $ (or other var starter) and then check if next char is { and previous is not escape \
        elif name[index] in var_start and next_char_is(name, index, "{") and not (index and name[index - 1] == "\\"):
            open_bracket = "{"
            close_bracket = "}"
            brackets += 1
            index += 1
        else:
            replaced += name[index]
        index += 1
    return replaced


def find_robot_vars(name: str) -> List[Tuple[int, int]]:
    """return list of tuples with (start, end) pos of vars in name"""
    var_start = set("$@%&")
    brackets = 0
    index = 0
    start = -1
    variables = []
    while index < len(name):
        if brackets:
            if name[index] == "{":
                brackets += 1
            elif name[index] == "}":
                brackets -= 1
                if not brackets:
                    variables.append((start, index + 1))
        # it looks for $ (or other var starter) and then check if next char is { and previous is not escape \
        elif name[index] in var_start and next_char_is(name, index, "{") and not (index and name[index - 1] == "\\"):
            brackets += 1
            start = index
            index += 1
        index += 1
    return variables


def pattern_type(value: str) -> Pattern:
    try:
        pattern = re.compile(value)
    except re.error as err:
        raise ValueError(f"Invalid regex pattern: {err}")
    return pattern


def get_section_name(node):
    if not node.header:
        #  Lines before first section are considered to be in `*** Comments ***` section even if header name is missing
        return "*** Comments ***"
    for token in node.header.data_tokens:
        if ROBOT_VERSION.major > 3:
            if token.type in node.header.handles_types:
                return token.value
        elif "HEADER" in token.type:
            return token.value
    return ""


def get_errors(node):
    if ROBOT_VERSION.major == 3:
        return [node.error] if node.error else []
    return node.errors



import helper
import sys
import helper
import helper
import helper
import helper
from helper import helper as sa
from helper.helper.helper import NULL
from helper.helper import NULL

Base = NULL()

def MyFunction():
	pass

class MyClassA(object):
	__tablename__ = 'tableA'
	n = sa.NULL(sa.NULL, primary_key=True)

class MyClassB(object):
	__tablename__ = 'tableB'
	n = sa.NULL(sa.NULL, primary_key=True)

class MyClassC(object):
	def __init__(self):
		pass

class MyClassD(object):
	def __init__(self):
		pass

if __name__ == "__main__":
	N = len(sys.argv)  # break here

	if N >= 2: # step 1
		arg1 = sys.argv[1]  # the debugger gets here even though N=1
		print(arg1)
	else:
		arg1 = 'MyString' # step 2

	if N >= 3: # step 3
		arg2 = int(sys.argv[2])  # the debugger gets here even though N=1
	else:
		arg2 = int(0) # step 4

	print(N)  # still N=1 step 5
	# print(arg1)  # if you print this then it changes the debugger behavior above


from tests_python import debugger_unittest
import sys
import re
import os

CHECK_BASELINE, CHECK_REGULAR, CHECK_CYTHON, CHECK_FRAME_EVAL = 'baseline', 'regular', 'cython', 'frame_eval'

pytest_plugins = [
    str('tests_python.debugger_fixtures'),
]

RUNS = 5


class PerformanceWriterThread(debugger_unittest.AbstractWriterThread):

    CHECK = None

    debugger_unittest.AbstractWriterThread.get_environ  # overrides

    def get_environ(self):
        env = os.environ.copy()
        if self.CHECK == CHECK_BASELINE:
            env['PYTHONPATH'] = r'X:\PyDev.Debugger.baseline'

        elif self.CHECK == CHECK_CYTHON:
            env['PYDEVD_USE_CYTHON'] = 'YES'
            env['PYDEVD_USE_FRAME_EVAL'] = 'NO'

        elif self.CHECK == CHECK_FRAME_EVAL:
            env['PYDEVD_USE_CYTHON'] = 'YES'
            env['PYDEVD_USE_FRAME_EVAL'] = 'YES'

        elif self.CHECK == CHECK_REGULAR:
            env['PYDEVD_USE_CYTHON'] = 'NO'
            env['PYDEVD_USE_FRAME_EVAL'] = 'NO'

        else:
            raise AssertionError("Don't know what to check.")
        return env

    debugger_unittest.AbstractWriterThread.get_pydevd_file  # overrides

    def get_pydevd_file(self):
        if self.CHECK == CHECK_BASELINE:
            return os.path.abspath(os.path.join(r'X:\PyDev.Debugger.baseline', 'pydevd.py'))
        dirname = os.path.dirname(__file__)
        dirname = os.path.dirname(dirname)
        return os.path.abspath(os.path.join(dirname, 'pydevd.py'))


class CheckDebuggerPerformance(debugger_unittest.DebuggerRunner):

    def get_command_line(self):
        return [sys.executable]

    def _get_time_from_result(self, stdout):
        match = re.search(r'TotalTime>>((\d|\.)+)<<', stdout)
        time_taken = match.group(1)
        return float(time_taken)

    def obtain_results(self, benchmark_name, filename):

        class PerformanceCheck(PerformanceWriterThread):
            TEST_FILE = debugger_unittest._get_debugger_test_file(filename)
            BENCHMARK_NAME = benchmark_name

        writer_thread_class = PerformanceCheck

        runs = RUNS
        all_times = []
        for _ in range(runs):
            stdout_ref = []

            def store_stdout(stdout, stderr):
                stdout_ref.append(stdout)

            with self.check_case(writer_thread_class) as writer:
                writer.additional_output_checks = store_stdout
                yield writer

            assert len(stdout_ref) == 1
            all_times.append(self._get_time_from_result(stdout_ref[0]))
            print('partial for: %s: %.3fs' % (writer_thread_class.BENCHMARK_NAME, all_times[-1]))
        if len(all_times) > 3:
            all_times.remove(min(all_times))
            all_times.remove(max(all_times))
        time_when_debugged = sum(all_times) / float(len(all_times))

        args = self.get_command_line()
        args.append(writer_thread_class.TEST_FILE)
        # regular_time = self._get_time_from_result(self.run_process(args, writer_thread=None))
        # simple_trace_time = self._get_time_from_result(self.run_process(args+['--regular-trace'], writer_thread=None))

        if 'SPEEDTIN_AUTHORIZATION_KEY' in os.environ:

            SPEEDTIN_AUTHORIZATION_KEY = os.environ['SPEEDTIN_AUTHORIZATION_KEY']

            # sys.path.append(r'X:\speedtin\pyspeedtin')
            import pyspeedtin  # If the authorization key is there, pyspeedtin must be available
            import pydevd
            pydevd_cython_project_id, pydevd_pure_python_project_id = 6, 7
            if writer_thread_class.CHECK == CHECK_BASELINE:
                project_ids = (pydevd_cython_project_id, pydevd_pure_python_project_id)
            elif writer_thread_class.CHECK == CHECK_REGULAR:
                project_ids = (pydevd_pure_python_project_id,)
            elif writer_thread_class.CHECK == CHECK_CYTHON:
                project_ids = (pydevd_cython_project_id,)
            else:
                raise AssertionError('Wrong check: %s' % (writer_thread_class.CHECK))
            for project_id in project_ids:
                api = pyspeedtin.PySpeedTinApi(authorization_key=SPEEDTIN_AUTHORIZATION_KEY, project_id=project_id)

                benchmark_name = writer_thread_class.BENCHMARK_NAME

                if writer_thread_class.CHECK == CHECK_BASELINE:
                    version = '0.0.1_baseline'
                    return  # No longer commit the baseline (it's immutable right now).
                else:
                    version = pydevd.__version__,

                commit_id, branch, commit_date = api.git_commit_id_branch_and_date_from_path(pydevd.__file__)
                api.add_benchmark(benchmark_name)
                api.add_measurement(
                    benchmark_name,
                    value=time_when_debugged,
                    version=version,
                    released=False,
                    branch=branch,
                    commit_id=commit_id,
                    commit_date=commit_date,
                )
                api.commit()

        self.performance_msg = '%s: %.3fs ' % (writer_thread_class.BENCHMARK_NAME, time_when_debugged)

    def method_calls_with_breakpoint(self):
        for writer in self.obtain_results('method_calls_with_breakpoint', '_performance_1.py'):
            writer.write_add_breakpoint(17, 'method')
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg

    def method_calls_without_breakpoint(self):
        for writer in self.obtain_results('method_calls_without_breakpoint', '_performance_1.py'):
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg

    def method_calls_with_step_over(self):
        for writer in self.obtain_results('method_calls_with_step_over', '_performance_1.py'):
            writer.write_add_breakpoint(26, None)

            writer.write_make_initial_run()
            hit = writer.wait_for_breakpoint_hit('111')

            writer.write_step_over(hit.thread_id)
            hit = writer.wait_for_breakpoint_hit('108')

            writer.write_run_thread(hit.thread_id)
            writer.finished_ok = True

        return self.performance_msg

    def method_calls_with_exception_breakpoint(self):
        for writer in self.obtain_results('method_calls_with_exception_breakpoint', '_performance_1.py'):
            writer.write_add_exception_breakpoint('ValueError')
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg

    def global_scope_1_with_breakpoint(self):
        for writer in self.obtain_results('global_scope_1_with_breakpoint', '_performance_2.py'):
            writer.write_add_breakpoint(writer.get_line_index_with_content('Breakpoint here'), None)
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg

    def global_scope_2_with_breakpoint(self):
        for writer in self.obtain_results('global_scope_2_with_breakpoint', '_performance_3.py'):
            writer.write_add_breakpoint(17, None)
            writer.write_make_initial_run()
            writer.finished_ok = True

        return self.performance_msg


if __name__ == '__main__':
    # Local times gotten (python 3.6)
    # method_calls_with_breakpoint: 1.150s
    # method_calls_without_breakpoint: 0.240s
    # method_calls_with_step_over: 2.680s
    # method_calls_with_exception_breakpoint: 0.235s
    # global_scope_1_with_breakpoint: 4.249s
    # global_scope_2_with_breakpoint: 2.807s
    # Checking: cython
    # method_calls_with_breakpoint: 0.526s
    # method_calls_without_breakpoint: 0.130s
    # method_calls_with_step_over: 1.133s
    # method_calls_with_exception_breakpoint: 0.136s
    # global_scope_1_with_breakpoint: 1.827s
    # global_scope_2_with_breakpoint: 1.405s
    # Checking: frame_eval
    # method_calls_with_breakpoint: 0.133s
    # method_calls_without_breakpoint: 0.128s
    # method_calls_with_step_over: 0.130s
    # method_calls_with_exception_breakpoint: 0.125s
    # global_scope_1_with_breakpoint: 0.281s
    # global_scope_2_with_breakpoint: 0.169s
    # TotalTime for profile: 209.01s

    debugger_unittest.SHOW_WRITES_AND_READS = False
    debugger_unittest.SHOW_OTHER_DEBUG_INFO = False
    debugger_unittest.SHOW_STDOUT = False

    import time
    start_time = time.time()

    tmpdir = None

    msgs = []
    for check in (
            # CHECK_BASELINE, -- Checks against the version checked out at X:\PyDev.Debugger.baseline.
            CHECK_REGULAR,
            CHECK_CYTHON,
            CHECK_FRAME_EVAL,
        ):
        PerformanceWriterThread.CHECK = check
        msgs.append('Checking: %s' % (check,))
        check_debugger_performance = CheckDebuggerPerformance(tmpdir)
        msgs.append(check_debugger_performance.method_calls_with_breakpoint())
        msgs.append(check_debugger_performance.method_calls_without_breakpoint())
        msgs.append(check_debugger_performance.method_calls_with_step_over())
        msgs.append(check_debugger_performance.method_calls_with_exception_breakpoint())
        msgs.append(check_debugger_performance.global_scope_1_with_breakpoint())
        msgs.append(check_debugger_performance.global_scope_2_with_breakpoint())

    for msg in msgs:
        print(msg)

    print('TotalTime for profile: %.2fs' % (time.time() - start_time,))



"""
Collection of classes for detecting checker disablers (like # robocop: disable) in robot files
"""
import re
from collections import defaultdict
from copy import deepcopy

from robot.utils import FileReader

import robocop.exceptions


class DisablersInFile:  # pylint: disable=too-few-public-methods
    """Container for file disablers"""

    def __init__(self):
        self.lastblock = -1
        self.lines = set()
        self.blocks = []

    def copy(self):
        """Used by defaultdict to create new instance for every new key in disablers container"""
        return deepcopy(self)


class DisablersFinder:
    """Parse all scanned file and find and disablers (in line or blocks)"""

    def __init__(self, filename, source=None):
        self.file_disabled = False
        self.any_disabler = False
        self.disabler_pattern = re.compile(r"robocop: ?(?P<disabler>disable|enable)=?(?P<rules>[\w\-,]*)")
        self.rules = defaultdict(DisablersInFile().copy)
        if source is not None:
            self._parse_source(source)
        else:
            self._parse_file(filename)

    def is_rule_disabled(self, rule_msg):
        """
        Check if given `rule_msg` is disabled. All takes precedence, then line disablers, then block disablers.
        We're checking for both message id and name.
        """
        if not self.any_disabler:
            return False
        return any(
            self.is_line_disabled(line, rule)
            for line in (rule_msg.line, *rule_msg.ext_disablers)
            for rule in ("all", rule_msg.rule_id, rule_msg.name)
        )

    def is_line_disabled(self, line, rule):
        """Helper method for is_rule_disabled that check if given line is in range of any disabled block"""
        if rule not in self.rules:
            return False
        if line in self.rules[rule].lines:
            return True
        return any(block[0] <= line <= block[1] for block in self.rules[rule].blocks)

    def _parse_lines(self, lines):
        lineno = -1
        for lineno, line in enumerate(lines, start=1):
            if "#" in line:
                self._parse_line(line, lineno)
        if lineno == -1:
            return
        self._end_block("all", lineno)
        self.file_disabled = self._is_file_disabled(lineno)
        self.any_disabler = len(self.rules) != 0

    def _parse_file(self, filename):
        try:
            with FileReader(filename) as file_reader:
                lines = list(file_reader.readlines())
                self._parse_lines(lines)
        except OSError:
            raise robocop.exceptions.FileError(filename) from None
        except UnicodeDecodeError:
            print(f"Failed to decode {filename}. Default supported encoding by Robot Framework is UTF-8. Skipping file")
            self.file_disabled = True

    def _parse_source(self, source):
        self._parse_lines(source.splitlines())

    def _parse_line(self, line, lineno):
        statement, comment = line.split("#", maxsplit=1)
        if "# noqa" in line:
            self._add_inline_disabler("all", lineno)
        disabler = self.disabler_pattern.search(comment)
        if not disabler:
            return
        if not disabler.group("rules"):
            rules = ["all"]
        else:
            rules = disabler.group("rules").split(",")
        block = not statement.lstrip()  # block disabler ignores preceding whitespaces
        if disabler.group("disabler") == "disable":
            for rule in rules:
                if block:
                    self._start_block(rule, lineno)
                else:
                    self._add_inline_disabler(rule, lineno)
        elif disabler.group("disabler") == "enable" and block:
            for rule in rules:
                self._end_block(rule, lineno)

    def _is_file_disabled(self, last_line):
        """
        The file is disabled if all rules are disabled in every lines - we need to iterate every block to see
        if they are linked from first to last line without breaks.
        """
        if "all" not in self.rules:
            return False
        prev_end = 1
        for block in self.rules["all"].blocks:
            if prev_end != block[0]:
                return False
            prev_end = block[1]
        return prev_end == last_line

    def _add_inline_disabler(self, rule, lineno):
        self.rules[rule].lines.add(lineno)

    def _start_block(self, rule, lineno):
        if self.rules[rule].lastblock == -1:
            self.rules[rule].lastblock = lineno

    def _end_block(self, rule, lineno):
        if rule == "all":
            self._end_all_blocks(lineno)
        if rule not in self.rules:
            return
        if self.rules[rule].lastblock != -1:
            block = (self.rules[rule].lastblock, lineno)
            self.rules[rule].lastblock = -1
            self.rules[rule].blocks.append(block)

    def _end_all_blocks(self, lineno):
        for rule in self.rules:
            if rule == "all":
                continue  # to avoid recursion
            self._end_block(rule, lineno)



def foo(arg):
    print('on foo mark', arg)
    return arg + 1


def main():
    items = [1]  # break here
    gen = (foo(arg) for arg in items)
    list(gen)

# import dis
# print('-------- main ------------')
# dis.dis(main)
# print('-------- foo ------------')
# dis.dis(foo)


if __name__ == '__main__':
    main()

    print('TEST SUCEEDED')



import os
import sys
import time
port = int(sys.argv[1])
root_dirname = os.path.dirname(os.path.dirname(__file__))

if root_dirname not in sys.path:
    sys.path.append(root_dirname)

import pydevd
try:
    pydevd._wait_for_attach()  # Cannot be called before _enable_attach.
except AssertionError:
    pass
else:
    raise AssertionError('Expected _wait_for_attach to raise exception.')

assert sys.gettrace() is None
print('enable attach to port: %s' % (port,))
pydevd._enable_attach(('127.0.0.1', port))
pydevd._enable_attach(('127.0.0.1', port))  # no-op in practice

try:
    pydevd._enable_attach(('127.0.0.1', port + 15))  # different port: raise error.
except AssertionError:
    pass
else:
    raise AssertionError('Expected _enable_attach to raise exception (because it is already hearing in another port).')

assert pydevd.get_global_debugger() is not None
assert sys.gettrace() is not None

a = 10  # Break 1
print('wait for attach')
pydevd._wait_for_attach()
print('finished wait for attach')
pydevd._wait_for_attach()  # Should promptly return (already connected).

a = 20  # Break 2

pydevd._wait_for_attach()  # As we disconnected on the 2nd break, this one should wait until a new configurationDone.

a = 20  # Break 3

while a == 20:  # Pause 1
    # The debugger should disconnect/reconnect, pause and then change 'a' to another value.
    time.sleep(1 / 20.)  # Pause 2

print('TEST SUCEEDED!')



# Original work Copyright 2012-2018 Matthew Tretter (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: port of https://github.com/matthewwithanm/python-markdownify using only
# the standard library (i.e.: html.parser).
import re
import logging
from functools import partial

log = logging.getLogger(__name__)

from html.parser import HTMLParser
from html.entities import name2codepoint

convert_heading_re = re.compile(r"convert_h(\d+)")
line_beginning_re = re.compile(r"^", re.MULTILINE)
whitespace_re = re.compile(r"[\r\n\s\t ]+")


def escape(text):
    if not text:
        return ""
    return (
        text.replace("_", r"\_")
        .replace("*", r"\*")
        .replace("[", r"\[")
        .replace("]", r"\]")
        .replace("#", r"\#")
    )


class _TagInfo(object):
    def __init__(self, tag, attrs):
        self.tag = tag
        self.attrs = attrs
        self.output = []


class _HTMLToMarkdownParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self.__output = []
        self.__tag = []
        self.__tag_info = []

        handlers = {}
        for d in dir(self):
            if d.startswith("_handle_"):
                handlers[d[8:]] = getattr(self, d)

        for i in range(10):
            handlers["h%s" % (i,)] = partial(self._handle_h, i)
        self._handlers = handlers

        # If we have something on start, check the start handlers too.
        handlers = {}
        for d in dir(self):
            if d.startswith("_start_handle_"):
                handlers[d[14:]] = getattr(self, d)

        self._start_handlers = handlers

    def handle_starttag(self, tag, attrs):
        # print("Start tag:", tag, type(tag))
        # for attr in attrs:
        #     print("     attr:", attr)

        if tag == "br":
            self._append("  \n")
            return

        elif tag in self._handlers:
            tag_info = _TagInfo(tag, attrs)
            self.__tag_info.append(tag_info)
            start_handler = self._start_handlers.get(tag)
            if start_handler is not None:
                start_handler(tag_info)

        self.__tag.append(tag)

    def handle_endtag(self, tag):
        # print("End tag  :", tag)

        if tag == "br":
            return

        else:
            handler = self._handlers.get(tag)
            if handler is not None:
                tag_info = self.__tag_info.pop()
                handler(tag_info)

        if self.__tag and self.__tag[-1] == tag:
            self.__tag.pop()

    def _extend(self, lst):
        output = self.__output
        if self.__tag_info:
            output = self.__tag_info[-1].output
        output.extend(lst)

    def _append(self, txt):
        output = self.__output
        if self.__tag_info:
            output = self.__tag_info[-1].output
        output.append(txt)

    def _handle_b(self, tag_info):
        if tag_info.output:
            self._append("**")
            self._extend(tag_info.output)
            self._append("**")

    _handle_strong = _handle_b

    def _handle_p(self, tag_info):
        self._extend(tag_info.output)
        self._append("\n\n")

    def _handle_a(self, tag_info):
        el = dict(tag_info.attrs)
        href = el.get("href")

        title = el.get("title")
        text = "".join(tag_info.output)

        # if text == href and not title:
        #     # Shortcut syntax
        #     self._append("<%s>" % href)
        # else:
        title_part = ' "%s"' % title.replace('"', r"\"") if title else ""

        if href.startswith("#"):
            if title:
                self._append("**%s**" % title.replace("*", r"\\*"))
            else:
                self._append("**%s**" % text.replace("*", r"\\*"))
            return

        self._append(
            "[%s](%s%s)" % (text or "", href, title_part) if href else text or ""
        )

    def _handle_img(self, tag_info):
        el = dict(tag_info.attrs)
        alt = el.get("alt", None) or ""
        src = el.get("src", None) or ""
        title = el.get("title", None) or ""
        title_part = ' "%s"' % title.replace('"', r"\"") if title else ""
        self._append("![")
        self._append(alt)
        self._append("](")
        self._append(src)
        self._append(title_part)
        self._append(")")

    def _handle_blockquote(self, tag_info):
        text = "".join(tag_info.output)

        self._append("\n")
        if text:
            self._append(line_beginning_re.sub("> ", text))

    def _handle_em(self, tag_info):
        if tag_info.output:
            self._append("*")
            self._extend(tag_info.output)
            self._append("*")

    _handle_i = _handle_em

    def _handle_table(self, tag_info):
        self._extend(tag_info.output)
        self._append("\n")

    def _handle_tr(self, tag_info):
        self._extend(tag_info.output)
        self._append("\n")

    def _handle_td(self, tag_info):
        self._extend(tag_info.output)
        self._append("\t")

    def _handle_h(self, n, tag_info):
        text = "".join(tag_info.output)
        text = text.rstrip()
        if n <= 2:
            line = "=" if n == 1 else "-"
            self._append(self.underline(text, line))
        else:
            hashes = "#" * n
            self._append("%s %s\n\n" % (hashes, text))

    def _handle_ol(self, tag_info):
        nested = False
        for parent_tag_info in self.__tag_info:
            if parent_tag_info.tag == "li":
                nested = True
        text = "".join(tag_info.output)
        if nested:
            text = "\n" + self.indent(text, 1)
        self._append(text)

    _handle_ul = _handle_ol

    def _handle_li(self, tag_info):
        try:
            parent_tag_info = self.__tag_info[-1]
        except:
            parent_tag_info = None

        if parent_tag_info is not None and parent_tag_info.tag == "ol":
            i_index_attr_name = "__i_index__"
            idx = getattr(parent_tag_info, i_index_attr_name, 0)
            bullet = "%s." % (idx + 1)
            setattr(parent_tag_info, i_index_attr_name, idx + 1)
        else:
            depth = -1
            for parent_tag_info in self.__tag_info:
                if parent_tag_info.tag == "ul":
                    depth += 1

            bullets = bullets = "*+-"
            try:
                bullet = bullets[depth % len(bullets)]
            except:
                bullet = "~"

        text = "".join(tag_info.output)
        self._append("%s %s\n" % (bullet, text))

    def indent(self, text, level):
        return line_beginning_re.sub("\t" * level, text) if text else ""

    def underline(self, text, pad_char):
        text = (text or "").rstrip()
        return "%s\n%s\n\n" % (text, pad_char * len(text)) if text else ""

    def handle_data(self, data):
        self._append(escape(whitespace_re.sub(" ", data or "")))
        # print("Data     :", data, type(data))

    def handle_comment(self, data):
        pass
        # print("Comment  :", data)

    def handle_entityref(self, name):
        try:
            c = chr(name2codepoint[name])
            self._append(c)
        except:
            log.exception("Error handling: %s", name)
        # print("Named ent:", c)

    def handle_charref(self, name):
        try:
            if name.startswith("x"):
                c = chr(int(name[1:], 16))
            else:
                c = chr(int(name))
            self._append(c)
        except:
            log.exception("Error handling: %s", name)
        # print("Num ent  :", c)

    def handle_decl(self, data):
        pass
        # print("Decl     :", data)

    def get_output(self):
        return "".join(self.__output)


def convert(html):
    parser = _HTMLToMarkdownParser()
    parser.feed(html)
    ret = parser.get_output()
    return ret



_robot_target_comm = None


def set_global_robot_target_comm(robot_target_comm):
    global _robot_target_comm
    _robot_target_comm = robot_target_comm


def get_global_robot_target_comm():
    return _robot_target_comm



import argparse
import sys
import os
import traceback

__file__ = os.path.abspath(__file__)
if __file__.endswith((".pyc", ".pyo")):
    __file__ = __file__[:-1]

LOG_FORMAT = "%(asctime)s UTC pid: %(process)d - %(threadName)s - %(levelname)s - %(name)s\n%(message)s\n\n"

_critical_error_log_file = os.path.join(
    os.path.expanduser("~"), "robotframework_interactive_critical.log"
)


def _critical_msg(msg):
    with open(_critical_error_log_file, "a+") as stream:
        stream.write(msg + "\n")


def add_arguments(parser):
    class DefaultOptions(object):
        verbose = 0

    parser.description = "RobotFramework Interactive"

    parser.add_argument(
        "--tcp", action="store_true", help="Use TCP client instead of stdio."
    )
    parser.add_argument(
        "--host",
        default="127.0.0.1",
        help="Connect to this IP address (i.e.: 127.0.0.01).",
    )
    parser.add_argument(
        "--port", type=int, default=-1, help="Connect to this port (i.e.: 1456)."
    )

    parser.add_argument(
        "--log-file",
        help="Redirect logs to the given file instead of writing to stderr (i.e.: c:/temp/my_log.log).",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=DefaultOptions.verbose,
        help="Increase verbosity of log output (i.e.: -vv).",
    )

    parser.add_argument(
        "--version",
        action="store_true",
        help="If passed, just prints the version to the standard output and exits.",
    )


def main(args=None, language_server_class=None):
    original_args = args if args is not None else sys.argv[1:]

    parser = argparse.ArgumentParser()
    add_arguments(parser)

    args = parser.parse_args(args=original_args)
    verbose = args.verbose
    log_file = args.log_file or ""

    try:
        try:
            import robotframework_interactive
        except ImportError:
            # Automatically add it to the path if __main__ is being executed.
            sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
            import robotframework_interactive  # @UnusedImport
        robotframework_interactive.import_robocorp_ls_core()

        from robocorp_ls_core.robotframework_log import (
            configure_logger,
            log_args_and_python,
            get_logger,
        )
    except:
        # Failed before having setup the logger (but after reading args).
        log_file = os.path.expanduser(log_file)
        log_file = os.path.realpath(os.path.abspath(log_file))
        dirname = os.path.dirname(log_file)
        basename = os.path.basename(log_file)
        try:
            os.makedirs(dirname)
        except:
            pass  # Ignore error if it already exists.

        name, ext = os.path.splitext(basename)
        postfix = "rf_interactive.init"
        log_file = os.path.join(
            dirname, name + "." + postfix + "." + str(os.getpid()) + ext
        )
        with open(log_file, "a+") as stream:
            traceback.print_exc(file=stream)

        raise

    if args.version:
        sys.stdout.write(robotframework_interactive.__version__)
        sys.stdout.flush()
        return

    configure_logger("rfinteractive", verbose, log_file)
    log = get_logger("robotframework_interactive.__main__")
    log_args_and_python(log, original_args, robotframework_interactive)

    try:
        from robotframework_interactive.server.options import Setup, Options

        Setup.options = Options(args)

        # Ok, now that we have our options, let's connect back.

        from robocorp_ls_core.python_ls import (
            start_io_lang_server,
            start_tcp_lang_client,
            binary_stdio,
        )

        if args.tcp:
            start_tcp_lang_client(args.host, args.port, language_server_class)
        else:
            stdin, stdout = binary_stdio()
            start_io_lang_server(stdin, stdout, language_server_class)

    except:
        log.exception("Error initializing")
        raise


if __name__ == "__main__":
    try:
        if sys.version_info[0] <= 2:
            raise AssertionError(
                "Python 3+ is required for RobotFramework Interactive.\nCurrent executable: "
                + sys.executable
            )
        main()
    except (SystemExit, KeyboardInterrupt):
        pass
    except:
        # Critical error (the logging may not be set up properly).
        # Print to file and stderr.
        with open(_critical_error_log_file, "a+") as stream:
            traceback.print_exc(file=stream)

        traceback.print_exc()



import subprocess
import sys
import os
env = os.environ.copy()
pythonpath = env.get('PYTHONPATH', '')
env['PYTHONPATH'] = os.path.dirname(__file__)


def call():
    print("called")  # break 1 here


if __name__ == '__main__':
    p = subprocess.Popen(
        [sys.executable, '-c', 'import _debugger_case_python_c;_debugger_case_python_c.call()'],
        stdout=subprocess.PIPE,
        env=env,
    )
    stdout, stderr = p.communicate()
    assert b'called' in stdout, 'Did not find b"called" in: %s' % (stdout,)
    print('TEST SUCEEDED!')  # break 2 here



from robocorp_ls_core.python_ls import PythonLanguageServer
from robocorp_ls_core.protocols import IConfig, ActionResultDict
from robocorp_ls_core.basic import overrides
from robocorp_ls_core.robotframework_log import get_logger
import threading
from robotframework_interactive.protocols import IMessage, IRobotFrameworkInterpreter
from robocorp_ls_core.options import DEFAULT_TIMEOUT, USE_TIMEOUTS, NO_TIMEOUT
from typing import Optional, Callable, Union
import sys
import traceback
import json

log = get_logger(__name__)


class StopControwFlowException(Exception):
    pass


class RfInterpreterServerApi(PythonLanguageServer):
    def __init__(self, read_from, write_to):
        PythonLanguageServer.__init__(self, read_from, write_to)
        from queue import Queue

        self._interpreter_queue: "Queue[IMessage]" = Queue()
        self._interpreter_initialized = False
        self._interpreter_disposed = False
        self._interpreter: Optional[IRobotFrameworkInterpreter]
        self._finished_event = threading.Event()

    @overrides(PythonLanguageServer._create_config)
    def _create_config(self) -> IConfig:
        from robocorp_ls_core.config import Config

        return Config()

    def m_interpreter__start(self, uri: str) -> ActionResultDict:
        if self._interpreter_initialized:
            return {
                "success": False,
                "message": "Interpreter already initialized",
                "result": None,
            }
        if self._interpreter_disposed:
            return {
                "success": False,
                "message": "Interpreter already disposed",
                "result": None,
            }

        try:
            from robotframework_interactive.interpreter import RobotFrameworkInterpreter

            interpreter = RobotFrameworkInterpreter()
            self._interpreter = interpreter

            def on_output(msg: str, category: str):
                if not self._interpreter_disposed:
                    endpoint = self._lsp_messages.endpoint
                    endpoint.notify(
                        "interpreter/output", {"output": msg, "category": category}
                    )

            def on_stdout(msg: str):
                on_output(msg, "stdout")

            def on_stderr(msg: str):
                on_output(msg, "stderr")

            def on_before_read():
                if not self._interpreter_disposed:
                    endpoint = self._lsp_messages.endpoint
                    endpoint.notify("interpreter/beforeRead", {})

            def on_after_read():
                if not self._interpreter_disposed:
                    endpoint = self._lsp_messages.endpoint
                    endpoint.notify("interpreter/afterRead", {})

            def on_exception_handled(e):
                err_as_string = traceback.format_exc()
                try:
                    classname = e.__class__.__name__
                except:
                    classname = str(type(e))

                on_output(
                    json.dumps(
                        {
                            "description": str(e),
                            "traceback": err_as_string,
                            "type": classname,
                        }
                    ),
                    "exception",
                )
                return True

            interpreter.on_stdout.register(on_stdout)
            interpreter.on_stderr.register(on_stderr)
            interpreter.on_exception_handled.register(on_exception_handled)
            interpreter.on_before_read.register(on_before_read)
            interpreter.on_after_read.register(on_after_read)

            started_main_loop_event = threading.Event()

            def run_on_thread():
                def on_main_loop(interpreter: IRobotFrameworkInterpreter):
                    import robot

                    info = {
                        "sys.version": sys.version,
                        "sys.executable": sys.executable,
                        "robot.version": robot.get_version(),
                        "initialization.finished": True,
                    }
                    on_output(json.dumps(info), "json_info")

                    started_main_loop_event.set()

                    # Ok, we'll be stopped at this point receiving events and
                    # processing them.
                    self._interpreter_initialized = True
                    while True:
                        msg: IMessage = self._interpreter_queue.get()
                        try:
                            msg(interpreter)
                        except StopControwFlowException as e:
                            break

                        except Exception as e:
                            log.exception("Error evaluating: %s", msg)
                            msg.action_result_dict = {
                                "success": False,
                                "message": str(e),
                                "result": None,
                            }
                        finally:
                            msg.event.set()

                try:
                    interpreter.initialize(on_main_loop)
                finally:
                    self._finished_event.set()
                    self._interpreter = None

            t = threading.Thread(target=run_on_thread)
            t.start()
            assert started_main_loop_event.wait(
                DEFAULT_TIMEOUT if USE_TIMEOUTS else None
            )

            # Ok, at this point it's initialized.

            return {"success": True, "message": None, "result": True}
        except Exception as e:
            log.exception()
            return {"success": False, "message": str(e), "result": None}

    def m_interpreter__evaluate(
        self, code: str
    ) -> Union[Callable[[], ActionResultDict], ActionResultDict]:
        if not self._interpreter_initialized:
            return {
                "success": False,
                "message": "Interpreter still not initialized.",
                "result": None,
            }
        if self._interpreter_disposed:
            return {
                "success": False,
                "message": "Interpreter already disposed",
                "result": None,
            }

        evaluate = _Evaluate(code)
        self._interpreter_queue.put(evaluate)

        def func():
            evaluate.event.wait()
            return evaluate.action_result_dict

        return func

    def m_interpreter__compute_evaluate_text(
        self, code: str, target_type: str
    ) -> ActionResultDict:
        """
        :param target_type:
            'evaluate': means that the target is an evaluation with the given code.
                This implies that the current code must be changed to make sense
                in the given context.

            'completions': means that the target is a code-completion
                This implies that the current code must be changed to include
                all previous evaluation so that the code-completion contains
                the full information up to the current point.
        """
        if not self._interpreter_initialized:
            return {
                "success": False,
                "message": "Interpreter still not initialized.",
                "result": None,
            }
        if self._interpreter_disposed:
            return {
                "success": False,
                "message": "Interpreter already disposed",
                "result": None,
            }

        interpreter = self._interpreter
        if not interpreter:
            return {"success": False, "message": "Interpreter is None", "result": None}

        evaluate_text = interpreter.compute_evaluate_text(code, target_type=target_type)
        return {"success": True, "message": None, "result": evaluate_text}

    def m_interpreter__stop(self) -> ActionResultDict:
        stop = _Stop()
        if not self._interpreter_initialized or self._interpreter_disposed:
            return {
                "success": True,
                "message": None,  # Do nothing if not initialized nor disposed.
                "result": None,
            }
        self._interpreter_queue.put(stop)
        stop.event.wait()
        self._finished_event.wait(DEFAULT_TIMEOUT if USE_TIMEOUTS else NO_TIMEOUT)
        return stop.action_result_dict


class _Evaluate(object):
    def __init__(self, code: str):
        self.code = code
        self.event = threading.Event()
        self.action_result_dict: ActionResultDict = {
            "success": False,
            "message": "Code not evaluated.",
            "result": None,
        }

    def __call__(self, interpreter: IRobotFrameworkInterpreter):
        result = interpreter.evaluate(self.code)
        self.action_result_dict = result


class _Stop(object):
    def __init__(self):
        self.event = threading.Event()
        self.action_result_dict: ActionResultDict = {
            "success": False,
            "message": "Stop not executed.",
            "result": None,
        }

    def __call__(self, interpreter: IRobotFrameworkInterpreter):
        self.action_result_dict = {"success": True, "message": None, "result": None}
        raise StopControwFlowException()



import threading
from typing import Optional, Any, Set, List, Dict, Iterable, Tuple, Iterator
import typing
import weakref

from robocorp_ls_core import uris
from robocorp_ls_core.basic import overrides
from robocorp_ls_core.cache import instance_cache
from robocorp_ls_core.lsp import (
    TextDocumentContentChangeEvent,
    TextDocumentItem,
    MarkupContentTypedDict,
    MarkupKind,
)
from robocorp_ls_core.protocols import (
    check_implements,
    IWorkspaceFolder,
    IEndPoint,
    ITestInfoFromSymbolsCacheTypedDict,
    ITestInfoFromUriTypedDict,
    IDocument,
    IConfig,
)
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.watchdog_wrapper import IFSObserver
from robocorp_ls_core.workspace import Workspace, Document
from robotframework_ls.constants import NULL
from robotframework_ls.impl._symbols_cache import BaseSymbolsCache
from robotframework_ls.impl.protocols import (
    IRobotWorkspace,
    IRobotDocument,
    IBaseCompletionContext,
    ISymbolsCache,
    ISymbolsJsonListEntry,
    ICompletionContext,
    IKeywordNode,
    ISymbolKeywordInfo,
    ICompletionContextWorkspaceCaches,
    IOnDependencyChanged,
)
from robotframework_ls.impl.robot_constants import ROBOT_FILE_EXTENSIONS


log = get_logger(__name__)


class _KeywordInfo:
    _documentation: MarkupContentTypedDict

    __slots__ = ["_node", "name", "_documentation"]

    def __init__(self, node: IKeywordNode):
        self._node = node
        self.name = node.name

    def get_documentation(self) -> MarkupContentTypedDict:
        from robotframework_ls.impl import ast_utils
        from robotframework_ls.impl.text_utilities import (
            build_keyword_docs_with_signature,
        )

        try:
            return self._documentation
        except AttributeError:
            docs = ast_utils.get_documentation_as_markdown(self._node)
            args = tuple(ast_utils.iter_keyword_arguments_as_str(self._node))

            docs = build_keyword_docs_with_signature(self.name, args, docs, "markdown")

            self._documentation = {
                "kind": MarkupKind.Markdown,
                "value": docs,
            }

        return self._documentation

    def __typecheckself__(self) -> None:
        _: ISymbolKeywordInfo = check_implements(self)


class _SymbolsCacheForAST(BaseSymbolsCache):
    _cached_keyword_info: List[ISymbolKeywordInfo]

    def __init__(self, *args, **kwargs):
        keywords = kwargs.pop("keywords")
        self._keywords: List[IKeywordNode] = keywords
        super(_SymbolsCacheForAST, self).__init__(*args, **kwargs)

    def iter_keyword_info(self) -> Iterator[ISymbolKeywordInfo]:
        try:
            yield from iter(self._cached_keyword_info)
        except:
            cache: List[ISymbolKeywordInfo] = []
            for k in self._keywords:
                keyword_info = _KeywordInfo(k)
                yield keyword_info
                cache.append(keyword_info)
            self._cached_keyword_info = cache

    def __typecheckself__(self) -> None:
        _: ISymbolsCache = check_implements(self)


def _compute_symbols_from_ast(completion_context: ICompletionContext) -> ISymbolsCache:
    from robotframework_ls.impl import ast_utils
    from robocorp_ls_core.lsp import SymbolKind
    from robotframework_ls.impl.text_utilities import normalize_robot_name
    from robotframework_ls.impl.code_lens import list_tests

    doc = completion_context.doc

    ast = completion_context.get_ast()
    symbols: List[ISymbolsJsonListEntry] = []
    uri = doc.uri

    keywords: List[IKeywordNode] = []
    for keyword_node_info in ast_utils.iter_keywords(ast):
        keywords.append(keyword_node_info.node)
        symbols.append(
            {
                "name": keyword_node_info.node.name,
                "kind": SymbolKind.Class,
                "location": {
                    "uri": uri,
                    "range": {
                        "start": {
                            "line": keyword_node_info.node.lineno - 1,
                            "character": keyword_node_info.node.col_offset,
                        },
                        "end": {
                            "line": keyword_node_info.node.end_lineno - 1,
                            "character": keyword_node_info.node.end_col_offset,
                        },
                    },
                },
                "containerName": doc.path,
            }
        )

    keywords_used = set()
    for keyword_usage_info in ast_utils.iter_keyword_usage_tokens(
        ast, collect_args_as_keywords=True
    ):
        keywords_used.add(normalize_robot_name(keyword_usage_info.name))

    test_info = list_tests(completion_context)
    test_info_for_cache: List[ITestInfoFromSymbolsCacheTypedDict] = [
        {"name": x["name"], "range": x["range"]} for x in test_info
    ]

    return _SymbolsCacheForAST(
        symbols,
        None,
        doc,
        keywords_used,
        uri=uri,
        test_info=test_info_for_cache,
        keywords=keywords,
    )


class _ReindexInfo(object):
    def __init__(self):
        self.uris_to_iter: Set[str] = set()
        self.full_reindex: bool = False
        self.finished_collection = threading.Event()


class _ReindexManager(object):
    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._disposed = False

        self._reindex_info = _ReindexInfo()
        self._reindex_event = threading.Event()

        # i.e.: start with a full reindex.
        self._reindex_info.full_reindex = True
        self._reindex_event.set()

    def request_uri_collection(self, doc_uri: str) -> _ReindexInfo:
        with self._lock:
            if not self._disposed:
                self._reindex_info.uris_to_iter.add(doc_uri)

            self._reindex_event.set()
            return self._reindex_info

    # This isn't ideal because we don't process file removals as we should...
    # def request_directory_collection(self, directory):
    #     with self._lock:
    #         if not self._disposed:
    #             try:
    #                 for entry in os.scandir(directory):
    #                     if entry.path.endswith((".robot", ".resource")):
    #                         self._reindex_info.uris_to_iter.add(
    #                             uris.from_fs_path(entry.path)
    #                         )
    #             except:
    #                 # Directory removed: request full collection
    #                 self._reindex_info.full_reindex = True
    #
    #         self._reindex_event.set()
    #         return self._reindex_info

    def request_full_collection(self) -> _ReindexInfo:
        with self._lock:
            if not self._disposed:
                self._reindex_info.full_reindex = True

            self._reindex_event.set()
            return self._reindex_info

    def wait_for_info_to_reindex(self) -> _ReindexInfo:
        self._reindex_event.wait()
        with self._lock:
            if self._disposed:
                return self._reindex_info

            self._reindex_event.clear()
            ret = self._reindex_info
            self._reindex_info = _ReindexInfo()
            return ret

    def dispose(self):
        with self._lock:
            self._disposed = True
            self._reindex_info.finished_collection.set()

            self._reindex_info = _ReindexInfo()
            self._reindex_info.finished_collection.set()

            self._reindex_event.set()


class WorkspaceIndexer(object):
    def __init__(
        self,
        robot_workspace,
        endpoint: Optional[IEndPoint],
        collect_tests: bool = False,
    ) -> None:
        self._robot_workspace = weakref.ref(robot_workspace)
        robot_workspace.on_file_changed.register(self._on_file_changed)
        self._endpoint = endpoint
        self._collect_tests = collect_tests
        self._clear_caches = threading.Event()
        self._reindex_manager = _ReindexManager()
        self._disposed = threading.Event()

        if collect_tests:
            assert endpoint is not None
        t = threading.Thread(target=self._on_thread)
        self._cached: Dict[str, Any] = {}
        t.daemon = True
        t.start()

    def _on_file_changed(self, filename: str):
        # with open("x:/temp/rara.txt", "a+") as stream:
        #     stream.write("%s\n" % filename)

        if filename and filename.endswith(ROBOT_FILE_EXTENSIONS):
            uri = uris.from_fs_path(filename)
            self._reindex_manager.request_uri_collection(uri)

    def wait_for_full_test_collection(self):
        assert (
            self._collect_tests
        ), "Cannot wait for first test collection if not collecting tests."

        self._clear_caches.set()
        reindex_info = self._reindex_manager.request_full_collection()
        reindex_info.finished_collection.wait()

        return True

    def _on_thread(self) -> None:
        if not self._collect_tests:
            for _uri, symbols_cache in self.iter_uri_and_symbols_cache():
                # Do a single collection at startup, afterwards only
                # collect again on demand.
                pass
        else:
            endpoint = self._endpoint
            assert endpoint
            while True:
                reindex_info = self._reindex_manager.wait_for_info_to_reindex()

                if self._disposed.is_set():
                    reindex_info.finished_collection.set()
                    return

                try:
                    if self._clear_caches.is_set() or reindex_info.full_reindex:
                        # If clear caches is set, we need to force the notifications.
                        force_notify = self._clear_caches.is_set()
                        self._clear_caches.clear()

                        old_cached = self._cached
                        new_cached = {}

                        test_info_lst: List[ITestInfoFromSymbolsCacheTypedDict]
                        for uri, symbols_cache in self.iter_uri_and_symbols_cache():
                            if symbols_cache is None:
                                test_info_lst = []
                            else:
                                lst = symbols_cache.get_test_info()
                                if lst is None:
                                    test_info_lst = []
                                else:
                                    test_info_lst = lst

                            if uri:
                                test_info_for_uri: ITestInfoFromUriTypedDict = {
                                    "uri": uri,
                                    "testInfo": test_info_lst,
                                }
                                new_cached[uri] = test_info_lst
                                if (
                                    old_cached.pop(uri, None) != test_info_lst
                                    or force_notify
                                ):
                                    endpoint.notify(
                                        "$/testsCollected",
                                        test_info_for_uri,
                                    )

                        for uri in old_cached.keys():
                            endpoint.notify(
                                "$/testsCollected",
                                {"uri": uri, "testInfo": []},
                            )
                        self._cached = new_cached
                    else:
                        # In this case we won't notify about old keys removed
                        # because they weren't found (although if uris_to_iter
                        # has removed uris, we'll still do the right thing).
                        cached = self._cached

                        uris_to_iter = reindex_info.uris_to_iter
                        for uri, symbols_cache in self.iter_uri_and_symbols_cache(
                            uris_to_iter=uris_to_iter
                        ):
                            if symbols_cache is None:
                                test_info_lst = []
                            else:
                                lst = symbols_cache.get_test_info()
                                if lst is None:
                                    test_info_lst = []
                                else:
                                    test_info_lst = lst

                            if uri:
                                test_info_for_uri = {
                                    "uri": uri,
                                    "testInfo": test_info_lst,
                                }
                                if cached.get(uri) != test_info_lst:
                                    cached[uri] = test_info_lst
                                    endpoint.notify(
                                        "$/testsCollected",
                                        test_info_for_uri,
                                    )
                finally:
                    reindex_info.finished_collection.set()

    def dispose(self):
        self._disposed.set()
        self._reindex_manager.dispose()

    def on_updated_document(self, doc_uri: str):
        self._reindex_manager.request_uri_collection(doc_uri)

    def on_updated_folders(self):
        self._reindex_manager.request_full_collection()

    def iter_uri_and_symbols_cache(
        self,
        only_for_open_docs=False,
        initial_time: Optional[float] = None,
        timeout: Optional[float] = None,
        context: Optional[IBaseCompletionContext] = None,
        found: Optional[Set[str]] = None,
        uris_to_iter: Optional[Set[str]] = None,
    ) -> Iterable[Tuple[str, Optional[ISymbolsCache]]]:
        from typing import cast
        import time

        if not found:
            found = set()

        if initial_time is None:
            initial_time = time.time()

        if timeout is None:
            timeout = 99999999

        workspace = self._robot_workspace()
        if not workspace:
            log.critical("self._robot_workspace already collected in WorkspaceIndexer.")
            return

        if uris_to_iter is not None:

            def iter_in():
                yield from iter(uris_to_iter)

        else:
            if only_for_open_docs:

                def iter_in():
                    for doc_uri in workspace.get_open_docs_uris():
                        yield doc_uri

            else:

                def iter_in():
                    for uri in workspace.iter_all_doc_uris_in_workspace(
                        ROBOT_FILE_EXTENSIONS
                    ):
                        yield uri

        for uri in iter_in():
            if not uri:
                continue

            if context is not None:
                context.check_cancelled()

            if time.time() - initial_time > timeout:
                log.info(
                    "Timed out gathering information from workspace symbols (only partial information was collected). Consider enabling the 'robot.workspaceSymbolsOnlyForOpenDocs' setting."
                )
                break

            if uri in found:
                continue
            found.add(uri)

            doc = cast(
                Optional[IRobotDocument],
                workspace.get_document(uri, accept_from_file=True),
            )
            if doc is None:
                yield uri, None  # i.e.: No longer there...
                continue

            symbols_cache = doc.symbols_cache
            if symbols_cache is None:
                from robotframework_ls.impl.completion_context import (
                    CompletionContext,
                )

                if context is not None:
                    ctx = CompletionContext(
                        doc,
                        monitor=context.monitor,
                        config=context.config,
                        workspace=workspace,
                    )
                else:
                    ctx = CompletionContext(
                        doc,
                        workspace=workspace,
                    )
                symbols_cache = _compute_symbols_from_ast(ctx)
            doc.symbols_cache = symbols_cache
            yield uri, symbols_cache


class RobotWorkspace(Workspace):
    def __init__(
        self,
        root_uri,
        fs_observer: IFSObserver,
        workspace_folders=None,
        libspec_manager=NULL,
        generate_ast=True,
        index_workspace=False,
        collect_tests=False,
        endpoint: Optional[IEndPoint] = None,
        on_dependency_changed: Optional[IOnDependencyChanged] = None,
    ):
        from robotframework_ls.impl.completion_context_workspace_caches import (
            CompletionContextWorkspaceCaches,
        )

        self.libspec_manager = libspec_manager

        # It needs to be set to None in the initialization (while we setup folders).
        self.workspace_indexer: Optional[WorkspaceIndexer] = None
        self.completion_context_workspace_caches: ICompletionContextWorkspaceCaches = (
            CompletionContextWorkspaceCaches(on_dependency_changed)
        )

        Workspace.__init__(
            self, root_uri, fs_observer, workspace_folders=workspace_folders
        )
        self._generate_ast = generate_ast
        if collect_tests:
            assert endpoint is not None
            assert index_workspace

        if index_workspace:
            self.workspace_indexer = WorkspaceIndexer(
                self, endpoint, collect_tests=collect_tests
            )
        else:
            self.workspace_indexer = None

        self.on_file_changed.register(
            self.completion_context_workspace_caches.on_file_changed
        )

    @overrides(Workspace.put_document)
    def put_document(self, text_document: TextDocumentItem) -> IDocument:
        doc = typing.cast(IRobotDocument, Workspace.put_document(self, text_document))
        self.completion_context_workspace_caches.on_updated_document(doc.uri, doc)
        if self.workspace_indexer is not None:
            self.workspace_indexer.on_updated_document(doc.uri)
        return doc

    @overrides(Workspace.update_document)
    def update_document(
        self, text_doc: TextDocumentItem, change: TextDocumentContentChangeEvent
    ) -> IDocument:
        doc = typing.cast(
            IRobotDocument, Workspace.update_document(self, text_doc, change)
        )
        self.completion_context_workspace_caches.on_updated_document(doc.uri, doc)
        if self.workspace_indexer is not None:
            self.workspace_indexer.on_updated_document(doc.uri)
        return doc

    @overrides(Workspace.remove_document)
    def remove_document(self, uri: str):
        doc = Workspace.remove_document(self, uri)
        self.completion_context_workspace_caches.on_updated_document(uri, None)
        if self.workspace_indexer is not None:
            self.workspace_indexer.on_updated_document(uri)
        return doc

    @overrides(Workspace.add_folder)
    def add_folder(self, folder: IWorkspaceFolder):
        Workspace.add_folder(self, folder)
        self.libspec_manager.add_workspace_folder(folder.uri)
        self.completion_context_workspace_caches.clear_caches()
        if self.workspace_indexer is not None:
            self.workspace_indexer.on_updated_folders()

    @overrides(Workspace.remove_folder)
    def remove_folder(self, folder_uri):
        Workspace.remove_folder(self, folder_uri)
        self.libspec_manager.remove_workspace_folder(folder_uri)
        self.completion_context_workspace_caches.clear_caches()
        if self.workspace_indexer is not None:
            self.workspace_indexer.on_updated_folders()

    @overrides(Workspace.on_changed_config)
    def on_changed_config(self, config: IConfig):
        Workspace.on_changed_config(self, config)
        self.completion_context_workspace_caches.clear_caches()

    @overrides(Workspace.dispose)
    def dispose(self):
        Workspace.dispose(self)
        self.completion_context_workspace_caches.dispose()
        indexer = self.workspace_indexer
        if indexer is not None:
            indexer.dispose()

    @overrides(Workspace._create_document)
    def _create_document(
        self, doc_uri, source=None, version=None, force_load_source=False
    ):
        return RobotDocument(
            doc_uri,
            source,
            version,
            generate_ast=self._generate_ast,
            mutate_thread=self._main_thread,
            force_load_source=force_load_source,
        )

    def __typecheckself__(self) -> None:
        _: IRobotWorkspace = check_implements(self)


class RobotDocument(Document):

    TYPE_TEST_CASE = "test_case"
    TYPE_INIT = "init"
    TYPE_RESOURCE = "resource"

    def __init__(
        self,
        uri,
        source=None,
        version=None,
        generate_ast=True,
        *,
        mutate_thread=None,
        force_load_source=False,
    ):
        Document.__init__(
            self,
            uri,
            source=source,
            version=version,
            mutate_thread=mutate_thread,
            force_load_source=force_load_source,
        )

        self._generate_ast = generate_ast
        self._ast = None
        self.symbols_cache = None

    @overrides(Document._clear_caches)
    def _clear_caches(self):
        Document._clear_caches(self)
        self._symbols_cache = None
        self.get_ast.cache_clear(self)  # noqa (clear the instance_cache).
        self.get_python_ast.cache_clear(self)  # noqa (clear the instance_cache).
        self.get_yaml_contents.cache_clear(self)  # noqa (clear the instance_cache).

    def get_type(self):
        path = self.path
        if not path:
            log.info("RobotDocument path empty.")
            return self.TYPE_TEST_CASE

        import os.path

        basename = os.path.basename(path)
        if basename.startswith("__init__"):
            return self.TYPE_INIT

        if basename.endswith(".resource"):
            return self.TYPE_RESOURCE

        return self.TYPE_TEST_CASE

    @instance_cache
    def get_ast(self):
        if not self._generate_ast:
            raise AssertionError(
                "The AST can only be accessed in the RobotFrameworkServerApi, not in the RobotFrameworkLanguageServer."
            )
        from robot.api import get_model, get_resource_model, get_init_model  # noqa

        try:
            source = self.source
        except:
            log.exception("Error getting source for: %s" % (self.uri,))
            source = ""

        try:
            t = self.get_type()
            if t == self.TYPE_TEST_CASE:
                ast = get_model(source)

            elif t == self.TYPE_RESOURCE:
                ast = get_resource_model(source)

            elif t == self.TYPE_INIT:
                ast = get_init_model(source)

            else:
                log.critical("Unrecognized section: %s", t)
                ast = get_model(source)

            ast.source = self.path
            return ast
        except:
            log.critical(f"Error parsing {self.uri}")
            # Note: we always want to return a valid AST here (the
            # AST itself should have the error).
            ast = get_model(f"*** Unable to parse: {self.uri} ***")
            ast.source = self.path
            return ast

    @instance_cache
    def get_python_ast(self):
        if not self._generate_ast:
            raise AssertionError(
                "The AST can only be accessed in the RobotFrameworkServerApi, not in the RobotFrameworkLanguageServer."
            )

        try:
            source = self.source
        except:
            log.exception("Error getting source for: %s" % (self.uri,))
            return None

        try:
            import ast as ast_module

            return ast_module.parse(source)
        except:
            log.critical(f"Error parsing python file: {self.uri}")
            return None

    @instance_cache
    def get_yaml_contents(self) -> Optional[Any]:
        try:
            source = self.source
        except:
            log.exception("Error getting source for: %s" % (self.uri,))
            return None

        try:
            from robocorp_ls_core import yaml_wrapper
            from io import StringIO

            s = StringIO()
            s.write(source)
            s.seek(0)
            return yaml_wrapper.load(s)
        except:
            log.critical(f"Error parsing yaml file: {self.uri}")
            return None

    def __typecheckself__(self) -> None:
        _: IRobotDocument = check_implements(self)



from _pydevd_bundle.pydevd_extension_api import StrPresentationProvider, TypeResolveProvider


class RectResolver(TypeResolveProvider):
    def get_dictionary(self, var):
        return {'length': var.length, 'width': var.width, 'area': var.length * var.width}

    def resolve(self, var, attribute):
        return getattr(var, attribute, None) if attribute != 'area' else var.length * var.width

    def can_provide(self, type_object, type_name):
        return type_name.endswith('Rect')


class RectToString(StrPresentationProvider):
    def get_str(self, val):
        return "Rectangle[Length: %s, Width: %s , Area: %s]" % (val.length, val.width, val.length * val.width)

    def can_provide(self, type_object, type_name):
        return type_name.endswith('Rect')



from functools import partial
import itertools
from pydevd import AbstractSingleNotificationBehavior
import time

import pytest

from _pydevd_bundle.pydevd_daemon_thread import run_as_pydevd_daemon_thread
from tests_python.debugger_unittest import CMD_THREAD_SUSPEND, CMD_STEP_OVER, CMD_SET_BREAK
from _pydev_bundle.pydev_override import overrides

STATE_RUN = 1
STATE_SUSPEND = 2


class _ThreadInfo(object):

    next_thread_id = partial(next, itertools.count())

    def __init__(self):
        self.state = STATE_RUN
        self.thread_id = self.next_thread_id()


class _CustomSingleNotificationBehavior(AbstractSingleNotificationBehavior):

    NOTIFY_OF_PAUSE_TIMEOUT = .01

    __slots__ = AbstractSingleNotificationBehavior.__slots__ + ['notification_queue']

    def __init__(self, py_db):
        try:
            from queue import Queue
        except ImportError:
            from Queue import Queue
        super(_CustomSingleNotificationBehavior, self).__init__(py_db)
        self.notification_queue = Queue()

    @overrides(AbstractSingleNotificationBehavior.send_resume_notification)
    def send_resume_notification(self, *args, **kwargs):
        # print('put resume', threading.current_thread())
        self.notification_queue.put('resume')

    @overrides(AbstractSingleNotificationBehavior.send_suspend_notification)
    def send_suspend_notification(self, *args, **kwargs):
        # print('put suspend', threading.current_thread())
        self.notification_queue.put('suspend')

    def do_wait_suspend(self, thread_info, stop_reason):
        with self.notify_thread_suspended(thread_info.thread_id, stop_reason=stop_reason):
            while thread_info.state == STATE_SUSPEND:
                time.sleep(.1)


@pytest.fixture
def _dummy_pydb():
    return _DummyPyDB()


@pytest.fixture(
    name='single_notification_behavior',
#     params=range(50)  #  uncomment to run the tests many times.
)
def _single_notification_behavior(_dummy_pydb):
    single_notification_behavior = _CustomSingleNotificationBehavior(_dummy_pydb)
    return single_notification_behavior


@pytest.fixture(name='notification_queue')
def _notification_queue(single_notification_behavior):
    return single_notification_behavior.notification_queue


def wait_for_notification(notification_queue, msg):
    __tracebackhide__ = True
    try:
        from Queue import Empty
    except ImportError:
        from queue import Empty
    try:
        found = notification_queue.get(timeout=2)
        assert found == msg
    except Empty:
        raise AssertionError('Timed out while waiting for %s notification.' % (msg,))


def join_thread(t):
    __tracebackhide__ = True
    t.join(2)
    assert not t.is_alive(), 'Thread still alive after timeout.s'


class _DummyPyDB(object):

    def __init__(self):
        self.created_pydb_daemon_threads = {}


def test_single_notification_1(single_notification_behavior, notification_queue):
    '''
    1. Resume before pausing 2nd thread

    - user pauses all (2) threads
    - break first -> send notification
    - user presses continue all before second is paused
      - 2nd should not pause nor send notification
      - resume all notification should be sent
    '''
    thread_info1 = _ThreadInfo()
    thread_info2 = _ThreadInfo()

    # pause all = set_suspend both
    single_notification_behavior.increment_suspend_time()
    single_notification_behavior.on_pause()
    thread_info1.state = STATE_SUSPEND
    thread_info2.state = STATE_SUSPEND

    dummy_py_db = _DummyPyDB()

    t = run_as_pydevd_daemon_thread(dummy_py_db, single_notification_behavior.do_wait_suspend, thread_info1, CMD_THREAD_SUSPEND)
    thread_info1.state = STATE_RUN
    # Set 2 to run before it starts (should not send additional message).
    thread_info2.state = STATE_RUN
    t.join()

    assert notification_queue.qsize() == 2
    assert notification_queue.get() == 'suspend'
    assert notification_queue.get() == 'resume'
    assert notification_queue.qsize() == 0

    # Run thread 2 only now (no additional notification).
    t = run_as_pydevd_daemon_thread(dummy_py_db, single_notification_behavior.do_wait_suspend, thread_info1, CMD_THREAD_SUSPEND)
    t.join()

    assert notification_queue.qsize() == 0


def test_single_notification_2(single_notification_behavior, notification_queue):
    '''
    2. Pausing all then stepping

    - user pauses all (2) threads
    - break first -> send notification
    - break second (no notification)
    - user steps in second
    - suspend in second -> send resume/pause notification on step
    '''
    thread_info1 = _ThreadInfo()
    thread_info2 = _ThreadInfo()

    dummy_py_db = _DummyPyDB()

    # pause all = set_suspend both
    single_notification_behavior.increment_suspend_time()
    single_notification_behavior.on_pause()
    thread_info1.state = STATE_SUSPEND
    thread_info2.state = STATE_SUSPEND

    # Leave both in break mode
    t1 = run_as_pydevd_daemon_thread(dummy_py_db, single_notification_behavior.do_wait_suspend, thread_info1, CMD_THREAD_SUSPEND)
    wait_for_notification(notification_queue, 'suspend')

    t2 = run_as_pydevd_daemon_thread(dummy_py_db, single_notification_behavior.do_wait_suspend, thread_info2, CMD_THREAD_SUSPEND)

    # Step would actually be set state to STEP, which would result in resuming
    # and then stopping again as if it was a SUSPEND (which calls a set_supend again with
    # the step mode).
    thread_info2.state = STATE_RUN
    join_thread(t2)
    wait_for_notification(notification_queue, 'resume')

    single_notification_behavior.increment_suspend_time()
    thread_info2.state = STATE_SUSPEND
    t2 = run_as_pydevd_daemon_thread(dummy_py_db, single_notification_behavior.do_wait_suspend, thread_info2, CMD_STEP_OVER)
    wait_for_notification(notification_queue, 'suspend')

    thread_info1.state = STATE_RUN
    thread_info2.state = STATE_RUN
    # First does a resume notification, the other remains quiet.
    wait_for_notification(notification_queue, 'resume')

    join_thread(t2)
    join_thread(t1)
    assert notification_queue.qsize() == 0


def test_single_notification_3(single_notification_behavior, notification_queue, _dummy_pydb):
    '''
    3. Deadlocked thread

    - user adds breakpoint in thread.join() -- just before threads becomes deadlocked
    - breakpoint hits -> send notification
      - pauses 2nd thread (no notification)
    - user steps over thead.join() -> never completes
    - user presses pause
      - second thread is already stopped
        - send notification regarding 2nd thread (still stopped).
    - leave both threads running: no suspend should be shown as there are no stopped threads
    - when thread is paused, show suspend notification
    '''

    # i.e.: stopping at breakpoint
    thread_info1 = _ThreadInfo()
    single_notification_behavior.increment_suspend_time()
    thread_info1.state = STATE_SUSPEND
    t1 = run_as_pydevd_daemon_thread(_dummy_pydb, single_notification_behavior.do_wait_suspend, thread_info1, CMD_SET_BREAK)

    # i.e.: stop because of breakpoint
    thread_info2 = _ThreadInfo()
    thread_info2.state = STATE_SUSPEND
    t2 = run_as_pydevd_daemon_thread(_dummy_pydb, single_notification_behavior.do_wait_suspend, thread_info2, CMD_SET_BREAK)

    wait_for_notification(notification_queue, 'suspend')

    # i.e.: step over (thread 2 is still suspended and this one never completes)
    thread_info1.state = STATE_RUN
    wait_for_notification(notification_queue, 'resume')

    join_thread(t1)

    # On pause we should notify that the thread 2 is suspended (after timeout if no other thread suspends first).
    single_notification_behavior.increment_suspend_time()
    single_notification_behavior.on_pause()
    thread_info1.state = STATE_SUSPEND
    thread_info2.state = STATE_SUSPEND
    wait_for_notification(notification_queue, 'suspend')

    thread_info2.state = STATE_RUN
    wait_for_notification(notification_queue, 'resume')

    join_thread(t2)
    assert notification_queue.qsize() == 0
    assert not single_notification_behavior._suspended_thread_ids

    # Now, no threads are running and pause is pressed
    # (maybe we could do a thread dump in this case as this
    # means nothing is stopped after pause is requested and
    # the timeout elapses).
    single_notification_behavior.increment_suspend_time()
    single_notification_behavior.on_pause()
    thread_info1.state = STATE_SUSPEND
    thread_info2.state = STATE_SUSPEND

    time.sleep(single_notification_behavior.NOTIFY_OF_PAUSE_TIMEOUT * 2)
    assert notification_queue.qsize() == 0

    t1 = run_as_pydevd_daemon_thread(_dummy_pydb, single_notification_behavior.do_wait_suspend, thread_info1, CMD_THREAD_SUSPEND)
    wait_for_notification(notification_queue, 'suspend')
    thread_info1.state = STATE_RUN
    wait_for_notification(notification_queue, 'resume')
    join_thread(t1)
    assert notification_queue.qsize() == 0


def test_single_notification_4(single_notification_behavior, notification_queue, _dummy_pydb):
    '''
    4. Delayed stop

    - user presses pause
    - stops first (2nd keeps running)
    - user steps on first
    - 2nd hits before first ends step (should not send any notification)
    - when step finishes send notification
    '''
    thread_info1 = _ThreadInfo()
    thread_info2 = _ThreadInfo()

    single_notification_behavior.increment_suspend_time()
    single_notification_behavior.on_pause()
    thread_info1.state = STATE_SUSPEND
    thread_info2.state = STATE_SUSPEND

    t1 = run_as_pydevd_daemon_thread(_dummy_pydb, single_notification_behavior.do_wait_suspend, thread_info1, CMD_THREAD_SUSPEND)
    wait_for_notification(notification_queue, 'suspend')
    thread_info1.state = STATE_RUN
    wait_for_notification(notification_queue, 'resume')
    join_thread(t1)

    t2 = run_as_pydevd_daemon_thread(_dummy_pydb, single_notification_behavior.do_wait_suspend, thread_info1, CMD_THREAD_SUSPEND)
    time.sleep(.1)
    assert notification_queue.qsize() == 0

    single_notification_behavior.increment_suspend_time()
    thread_info1.state = STATE_SUSPEND
    t1 = run_as_pydevd_daemon_thread(_dummy_pydb, single_notification_behavior.do_wait_suspend, thread_info1, CMD_STEP_OVER)
    wait_for_notification(notification_queue, 'suspend')
    thread_info2.state = STATE_RUN
    thread_info1.state = STATE_RUN
    join_thread(t1)
    join_thread(t2)
    wait_for_notification(notification_queue, 'resume')
    assert notification_queue.qsize() == 0




import ast

from robot.api.parsing import ModelTransformer, Token

from robotidy.decorators import check_start_end_line
from robotidy.exceptions import InvalidParameterValueError


# TODO: preserve comments?
class RemoveEmptySettings(ModelTransformer):
    """
    Remove empty settings.

    You can configure which settings are affected by parameter ``work_mode``. Possible values:
        - overwrite_ok (default): does not remove settings that are overwriting suite settings (Test Setup,
          Test Teardown, Test Template, Test Timeout or Default Tags)
        - always : works on every settings

    Empty settings that are overwriting suite settings will be converted to be more explicit
    (given that there is related suite settings present)::

        No timeout
        [Documentation]    Empty timeout means no timeout even when Test Timeout has been used.
        [Timeout]

    To::

        No timeout
        [Documentation]    Disabling timeout with NONE works too and is more explicit.
        [Timeout]    NONE

    You can disable that behavior by changing ``more_explicit`` parameter value to ``False``.

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/RemoveEmptySettings.html for more examples.
    """

    def __init__(self, work_mode: str = "overwrite_ok", more_explicit: bool = True):
        if work_mode not in ("overwrite_ok", "always"):
            raise InvalidParameterValueError(
                self.__class__.__name__, "work_mode", work_mode, "Possible values:\n    overwrite_ok\n    always"
            )
        self.work_mode = work_mode
        self.more_explicit = more_explicit
        self.overwritten_settings = set()
        self.child_types = {
            Token.SETUP,
            Token.TEARDOWN,
            Token.TIMEOUT,
            Token.TEMPLATE,
            Token.TAGS,
        }

    @check_start_end_line
    def visit_Statement(self, node):  # noqa
        # when not setting type or setting type but not empty
        if node.type not in Token.SETTING_TOKENS or len(node.data_tokens) != 1:
            return node
        # when empty and not overwriting anything - remove
        if (
            node.type not in self.child_types
            or self.work_mode == "always"
            or node.type not in self.overwritten_settings
        ):
            return None
        if self.more_explicit:
            indent = node.tokens[0].value if node.tokens[0].type == Token.SEPARATOR else ""
            setting_token = node.data_tokens[0]
            node.tokens = [
                Token(Token.SEPARATOR, indent),
                setting_token,
                Token(Token.SEPARATOR, self.formatting_config.separator),
                Token(Token.ARGUMENT, "NONE"),
                Token(Token.EOL, "\n"),
            ]
        return node

    def visit_File(self, node):  # noqa
        if self.work_mode == "overwrite_ok":
            self.overwritten_settings = self.find_overwritten_settings(node)
        self.generic_visit(node)
        self.overwritten_settings = set()

    @staticmethod
    def find_overwritten_settings(node):
        auto_detector = FindSuiteSettings()
        auto_detector.visit(node)
        return auto_detector.suite_settings


class FindSuiteSettings(ast.NodeVisitor):
    def __init__(self):
        self.suite_settings = set()

    def check_setting(self, node, overwritten_type):
        if len(node.data_tokens) != 1:
            self.suite_settings.add(overwritten_type)

    def visit_TestSetup(self, node):  # noqa
        self.check_setting(node, Token.SETUP)

    def visit_TestTeardown(self, node):  # noqa
        self.check_setting(node, Token.TEARDOWN)

    def visit_TestTemplate(self, node):  # noqa
        self.check_setting(node, Token.TEMPLATE)

    def visit_TestTimeout(self, node):  # noqa
        self.check_setting(node, Token.TIMEOUT)

    def visit_DefaultTags(self, node):  # noqa
        self.check_setting(node, Token.TAGS)



'''
This module was created to get information available in the interpreter, such as libraries,
paths, etc.

what is what:
sys.builtin_module_names: contains the builtin modules embeeded in python (rigth now, we specify all manually).
sys.prefix: A string giving the site-specific directory prefix where the platform independent Python files are installed

format is something as
EXECUTABLE:python.exe|libs@compiled_dlls$builtin_mods

all internal are separated by |
'''
import sys

try:
    import os.path

    def fully_normalize_path(path):
        '''fixes the path so that the format of the path really reflects the directories in the system
        '''
        return os.path.normpath(path)

    join = os.path.join
except:  # ImportError or AttributeError.

    # See: http://stackoverflow.com/questions/10254353/error-while-installing-jython-for-pydev
    def fully_normalize_path(path):
        '''fixes the path so that the format of the path really reflects the directories in the system
        '''
        return path

    def join(a, b):
        if a.endswith('/') or a.endswith('\\'):
            return a + b
        return a + '/' + b

IS_PYTHON_3_ONWARDS = 0

try:
    IS_PYTHON_3_ONWARDS = sys.version_info[0] >= 3
except:
    # That's OK, not all versions of python have sys.version_info
    pass

try:
    # Just check if False and True are defined (depends on version, not whether it's jython/python)
    False
    True
except:
    exec ('True, False = 1,0')  # An exec is used so that python 3k does not give a syntax error

if sys.platform == "cygwin":

    import ctypes

    def native_path(path):
        MAX_PATH = 512  # On cygwin NT, its 260 lately, but just need BIG ENOUGH buffer
        '''Get the native form of the path, like c:\\Foo for /cygdrive/c/Foo'''

        retval = ctypes.create_string_buffer(MAX_PATH)
        path = fully_normalize_path(path)
        path = tobytes(path)
        CCP_POSIX_TO_WIN_A = 0
        cygwin1dll = ctypes.cdll.LoadLibrary('cygwin1.dll')
        cygwin1dll.cygwin_conv_path(CCP_POSIX_TO_WIN_A, path, retval, MAX_PATH)

        return retval.value

else:

    def native_path(path):
        return fully_normalize_path(path)


def __getfilesystemencoding():
    '''
    Note: there's a copy of this method in _pydev_filesystem_encoding.py
    '''
    try:
        ret = sys.getfilesystemencoding()
        if not ret:
            raise RuntimeError('Unable to get encoding.')
        return ret
    except:
        try:
            # Handle Jython
            from java.lang import System  # @UnresolvedImport
            env = System.getProperty("os.name").lower()
            if env.find('win') != -1:
                return 'ISO-8859-1'  # mbcs does not work on Jython, so, use a (hopefully) suitable replacement
            return 'utf-8'
        except:
            pass

        # Only available from 2.3 onwards.
        if sys.platform == 'win32':
            return 'mbcs'
        return 'utf-8'


def getfilesystemencoding():
    try:
        ret = __getfilesystemencoding()

        # Check if the encoding is actually there to be used!
        if hasattr('', 'encode'):
            ''.encode(ret)
        if hasattr('', 'decode'):
            ''.decode(ret)

        return ret
    except:
        return 'utf-8'


file_system_encoding = getfilesystemencoding()

if IS_PYTHON_3_ONWARDS:
    unicode_type = str
    bytes_type = bytes

else:
    unicode_type = unicode
    bytes_type = str


def tounicode(s):
    if hasattr(s, 'decode'):
        if not isinstance(s, unicode_type):
            # Depending on the platform variant we may have decode on string or not.
            return s.decode(file_system_encoding)
    return s


def tobytes(s):
    if hasattr(s, 'encode'):
        if not isinstance(s, bytes_type):
            return s.encode(file_system_encoding)
    return s


def toasciimxl(s):
    # output for xml without a declared encoding

    # As the output is xml, we have to encode chars (< and > are ok as they're not accepted in the filesystem name --
    # if it was allowed, we'd have to do things more selectively so that < and > don't get wrongly replaced).
    s = s.replace("&", "&amp;")

    try:
        ret = s.encode('ascii', 'xmlcharrefreplace')
    except:
        # use workaround
        ret = ''
        for c in s:
            try:
                ret += c.encode('ascii')
            except:
                try:
                    # Python 2: unicode is a valid identifier
                    ret += unicode("&#%d;") % ord(c)
                except:
                    # Python 3: a string is already unicode, so, just doing it directly should work.
                    ret += "&#%d;" % ord(c)
    return ret


if __name__ == '__main__':
    try:
        # just give some time to get the reading threads attached (just in case)
        import time
        time.sleep(0.1)
    except:
        pass

    try:
        executable = tounicode(native_path(sys.executable))
    except:
        executable = tounicode(sys.executable)

    if sys.platform == "cygwin" and not executable.endswith(tounicode('.exe')):
        executable += tounicode('.exe')

    try:
        major = str(sys.version_info[0])
        minor = str(sys.version_info[1])
    except AttributeError:
        # older versions of python don't have version_info
        import string
        s = string.split(sys.version, ' ')[0]
        s = string.split(s, '.')
        major = s[0]
        minor = s[1]

    s = tounicode('%s.%s') % (tounicode(major), tounicode(minor))

    contents = [tounicode('<xml>')]
    contents.append(tounicode('<version>%s</version>') % (tounicode(s),))

    contents.append(tounicode('<executable>%s</executable>') % tounicode(executable))

    # this is the new implementation to get the system folders
    # (still need to check if it works in linux)
    # (previously, we were getting the executable dir, but that is not always correct...)
    prefix = tounicode(native_path(sys.prefix))
    # print_ 'prefix is', prefix

    result = []

    path_used = sys.path
    try:
        path_used = path_used[1:]  # Use a copy (and don't include the directory of this script as a path.)
    except:
        pass  # just ignore it...

    for p in path_used:
        p = tounicode(native_path(p))

        try:
            import string  # to be compatible with older versions
            if string.find(p, prefix) == 0:  # was startswith
                result.append((p, True))
            else:
                result.append((p, False))
        except (ImportError, AttributeError):
            # python 3k also does not have it
            # jython may not have it (depending on how are things configured)
            if p.startswith(prefix):  # was startswith
                result.append((p, True))
            else:
                result.append((p, False))

    for p, b in result:
        if b:
            contents.append(tounicode('<lib path="ins">%s</lib>') % (p,))
        else:
            contents.append(tounicode('<lib path="out">%s</lib>') % (p,))

    # no compiled libs
    # nor forced libs

    for builtinMod in sys.builtin_module_names:
        contents.append(tounicode('<forced_lib>%s</forced_lib>') % tounicode(builtinMod))

    contents.append(tounicode('</xml>'))
    unic = tounicode('\n').join(contents)
    inasciixml = toasciimxl(unic)
    if IS_PYTHON_3_ONWARDS:
        # This is the 'official' way of writing binary output in Py3K (see: http://bugs.python.org/issue4571)
        sys.stdout.buffer.write(inasciixml)
    else:
        sys.stdout.write(inasciixml)

    sys.stdout.flush()
    sys.stderr.flush()



import threading


class KeepInLoop(object):
    keep_in_loop = True  # Debugger should change to False to break.


def stop_loop():
    KeepInLoop.keep_in_loop = False
    return 'stopped_loop'


def double_number(number):
    while KeepInLoop.keep_in_loop:
        doubled = number * 2
        print(doubled)
        import time
        time.sleep(.5)
    return doubled


if __name__ == '__main__':
    threads = []
    for num in range(2):
        thread = threading.Thread(target=double_number, args=(num,))
        thread.start()
        threads.append(thread)

    for thread in threads:
        thread.join()

    print('TEST SUCEEDED!')



import trio


async def count():
    print('enter count')
    await trio.sleep(.001)  # break count 1
    await trio.sleep(.001)  # break count 2


async def count2():
    print('enter count 2')
    await trio.sleep(.001)
    await trio.sleep(.001)


async def count3():
    print('enter count 3')
    await trio.sleep(.001)
    await trio.sleep(.001)


async def main():
    await count()  # break main
    await count2()  # step main
    await count3()


if __name__ == '__main__':
    trio.run(main)
    print('TEST SUCEEDED!')



import typing
from typing import Dict, Iterator

T = typing.TypeVar("T")


class OrderedSet(typing.Generic[T]):
    """
    A simple ordered set with a regular dict as a basis.
    """

    def __init__(self, initial=None):
        self._dct: Dict[T, bool] = dict.fromkeys(initial or ())

    def add(self, obj: T) -> None:
        self._dct[obj] = True

    def discard(self, obj: T) -> None:
        self._dct.pop(obj, None)

    def __iter__(self) -> Iterator[T]:
        return iter(self._dct)

    def __contains__(self, obj: object) -> bool:
        return obj in self._dct

    def __bool__(self) -> bool:
        return bool(self._dct)

    def __len__(self) -> int:
        return len(self._dct)

    def __repr__(self) -> str:
        data = repr(list(self._dct)) if self._dct else ""
        return f"OrderedSet({data})"

    __str__ = __repr__



from typing import Set
import os
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.text_utilities import normalize_robot_name
from robocorp_ls_core.options import is_true_in_env

log = get_logger(__name__)


class IgnoreFailuresInStack:
    """
    We load the contents from environment variables:

    RFLS_IGNORE_FAILURES_IN_KEYWORDS:
        A (json-formatted) list of keywords where failures should be ignored.
        Note: ignored means they won't be reported as errors and the debugger
        won't break on them.

    The list below is always ignored by default (so using RFLS_IGNORE_FAILURES_IN_KEYWORDS
    it's possible to add other items to that list).

    [
        "run keyword and continue on failure",
        "run keyword and expect error",
        "run keyword and ignore error",
        "run keyword and warn on failure",
        "wait until keyword succeeds"
    ]

    It's also possible to set `RFLS_IGNORE_FAILURES_IN_KEYWORDS_OVERRIDE=1` to provide
    all the items if one of those shouldn't be there.
    """

    def __init__(self):
        import json
        from collections import deque

        self._stack: "Deque[str]" = deque()
        self.ignore_failures_inside: Set[str] = set()

        # Add default excludes.
        for entry in (
            "run keyword and continue on failure",
            "run keyword and expect error",
            "run keyword and ignore error",
            "run keyword and warn on failure",
            "wait until keyword succeeds",
            "run keyword and return status",
        ):
            self.ignore_failures_inside.add(normalize_robot_name(entry))

        if is_true_in_env("RFLS_IGNORE_FAILURES_IN_KEYWORDS_OVERRIDE"):
            self.ignore_failures_inside.clear()

        # Load additional excludes from the environment.
        ignore_failures_inside_in_env = os.getenv("RFLS_IGNORE_FAILURES_IN_KEYWORDS")
        if ignore_failures_inside_in_env:
            try:
                loaded = json.loads(ignore_failures_inside_in_env)
            except:
                log.exception(
                    "Error: unable to load RFLS_IGNORE_FAILURES_IN_KEYWORDS (%s) as a json.",
                    ignore_failures_inside_in_env,
                )
            else:
                if not isinstance(loaded, list):
                    log.critical(
                        "Expected RFLS_IGNORE_FAILURES_IN_KEYWORDS to be a json list of strings. Found: %s",
                        type(loaded),
                    )

                else:

                    for entry in loaded:
                        self.ignore_failures_inside.add(normalize_robot_name(entry))

    def ignore(self) -> bool:
        for name in self._stack:
            normalized = normalize_robot_name(name)
            if normalized in self.ignore_failures_inside:
                return True
        return False

    def push(self, name: str):
        self._stack.append(name)

    def pop(self):
        try:
            self._stack.pop()
        except:
            log.exception("Error in IgnoreFailuresInStack.pop()")



import threading

from robocorp_ls_core.protocols import IEndPoint, IDirCache, IProgressReporter
from contextlib import contextmanager
from typing import Optional, Iterator
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.basic import implements


log = get_logger(__name__)


def _next_id():
    # Note: changed to uuid from incremental because multiple processes
    # may start the progress and it shouldn't conflict from one to the
    # other.
    import uuid

    return str(uuid.uuid4())


class _ProgressReporter(object):

    _MIN_TIME = 0.25

    def __init__(
        self,
        endpoint: IEndPoint,
        title: str,
        dir_cache: Optional[IDirCache],
        elapsed_time_key=None,
    ) -> None:
        from robocorp_ls_core.timeouts import TimeoutTracker
        import time

        self.endpoint = endpoint
        self.title = title

        self._started = False
        self._finished = False

        self._lock = threading.Lock()
        self._id = _next_id()

        self._expected_time = None
        self._initial_time = time.time()

        self._additional_info: str = ""

        self._dir_cache = dir_cache
        self._last_elapsed_time_key = (
            elapsed_time_key
            if elapsed_time_key is not None
            else ("operation_time", title)
        )
        try:
            if dir_cache:
                expected_time = dir_cache.load(self._last_elapsed_time_key, float)
                # Leave some gap on the expected.
                self._expected_time = expected_time * 1.2
        except KeyError:
            pass

        self._last_progress = 0.0
        self.timeout_tracker = TimeoutTracker.get_singleton()
        self.timeout_tracker.call_on_timeout(self._MIN_TIME, self._on_first_timeout)

    def _on_first_timeout(self):
        with self._lock:
            if not self._finished and not self._started:
                self._started = True
                self.endpoint.notify(
                    "$/customProgress",
                    {"kind": "begin", "id": self._id, "title": self.title},
                )
                if self._expected_time:
                    update_time = self._expected_time / 30.0
                else:
                    update_time = 0.25

                self.timeout_tracker.call_on_timeout(
                    update_time, self._on_recurrent_timeout
                )

    def _on_recurrent_timeout(self) -> None:
        import time

        with self._lock:
            if not self._finished and self._started:
                elapsed_time = time.time() - self._initial_time
                expected_time = self._expected_time

                if not self._additional_info:
                    msg = "Elapsed: %.1fs" % (elapsed_time,)
                else:
                    msg = "Elapsed: %.1fs : %s" % (
                        elapsed_time,
                        self._additional_info,
                    )

                args = {
                    "kind": "report",
                    "id": self._id,
                    "message": msg,
                }
                if expected_time:
                    progress = elapsed_time / expected_time
                    if progress > 0.95:
                        progress = 0.95
                    increment = (progress - self._last_progress) * 100
                    self._last_progress = progress
                    args["increment"] = increment

                self.endpoint.notify("$/customProgress", args)
                self.timeout_tracker.call_on_timeout(0.5, self._on_recurrent_timeout)

    @implements(IProgressReporter.set_additional_info)
    def set_additional_info(self, additional_info: str) -> None:
        self._additional_info = additional_info

    def finish(self) -> None:
        import time

        with self._lock:
            if not self._finished:
                self._finished = True
                self.endpoint.notify(
                    "$/customProgress", {"kind": "end", "id": self._id}
                )
                total_elapsed_time = time.time() - self._initial_time
                if total_elapsed_time > self._MIN_TIME:
                    dir_cache = self._dir_cache
                    if dir_cache:
                        dir_cache.store(self._last_elapsed_time_key, total_elapsed_time)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IProgressReporter = check_implements(self)


_progress_context = threading.local()


def get_current_progress_reporter() -> Optional[_ProgressReporter]:
    try:
        try:
            stack = _progress_context._stack
        except AttributeError:
            return None
        else:
            try:
                return stack[-1]
            except IndexError:
                return None
    except Exception:
        log.exception("Unexpected error getting current progress reporter.")
        return None


class ProgressWrapperForTotalWork:
    """
    Wraps an IProgressReporter to have a quick way to show stes/total steps.

    i.e.:

    with progress_context(...) as progress_reporter:
        progress_wrapper = ProgressWrapperForTotalWork(progress_reporter)

        # Schedule many steps and at each point call.
        progress_reporter.increment_total_steps()

        # When a step is done, increment steps done.
        progress_reporter.increment_step_done()
    """

    def __init__(
        self,
        progress_reporter: IProgressReporter,
        message: str = "%s of %s",
    ) -> None:
        self.progress_reporter = progress_reporter
        self.message = message
        self._lock = threading.Lock()
        self._total_steps = 0
        self._current_step = 0

    def increment_total_steps(self):
        with self._lock:
            self._total_steps += 1
            self._update_message()

    def increment_step_done(self):
        with self._lock:
            self._current_step += 1
            self._update_message()

    def _update_message(self):
        self.progress_reporter.set_additional_info(
            self.message % (self._current_step, self._total_steps)
        )


@contextmanager
def progress_context(
    endpoint: IEndPoint,
    title: str,
    dir_cache: Optional[IDirCache],
    elapsed_time_key=None,
) -> Iterator[IProgressReporter]:
    """
    Creates a progress context which submits $/customProgress notifications to the
    client.

    Automatically updates the progress based on a previous invocation for some
    action with the same title (stores the elapsed time at the dir_cache).

    :param dir_cache:
        If None, an estimate for the task is not loaded/saved.

    :param elapsed_time_key:
        If None, the default is using the title as an entry in the dir cache,
        otherwise, the given key is used to load/save the time taken in the
        cache dir.
    """
    progress_reporter = _ProgressReporter(
        endpoint, title, dir_cache, elapsed_time_key=elapsed_time_key
    )
    try:
        stack = _progress_context._stack
    except AttributeError:
        stack = _progress_context._stack = []

    stack.append(progress_reporter)
    try:
        yield progress_reporter
    finally:
        del stack[-1]
        progress_reporter.finish()



in_global_scope = 'in_global_scope_value'


class SomeClass(object):

    def method(self):
        print('breakpoint here')


if __name__ == '__main__':
    SomeClass().method()
    print('TEST SUCEEDED')



from functools import lru_cache
from pathlib import Path

from pathspec import PathSpec

from robocop.exceptions import FileError


DEFAULT_EXCLUDES = r"(\.direnv|\.eggs|\.git|\.hg|\.nox|\.tox|\.venv|venv|\.svn)"


def find_project_root(srcs):
    """Return a directory containing .git, .robocop or pyproject.toml.
    That directory will be a common parent of all files and directories
    passed in `srcs`.
    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.
    """
    if not srcs:
        return Path("/").resolve()

    path_srcs = [Path(Path.cwd(), src).resolve() for src in srcs]

    # A list of lists of parents for each 'src'. 'src' is included as a
    # "parent" of itself if it is a directory
    src_parents = [list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs]

    common_base = max(
        set.intersection(*(set(parents) for parents in src_parents)),
        key=lambda path: path.parts,
    )

    for directory in (common_base, *common_base.parents):
        if (
            (directory / ".git").exists()
            or (directory / "pyproject.toml").is_file()
            or (directory / ".robocop").is_file()
        ):
            return directory
    return directory


def find_file_in_project_root(config_name, root):
    for parent in (root, *root.parents):
        if (parent / ".git").exists() or (parent / config_name).is_file():
            return parent / config_name
    return parent / config_name


@lru_cache()
def get_gitignore(root):
    """Return a PathSpec matching gitignore content if present."""
    gitignore = root / ".gitignore"
    lines = []
    if gitignore.is_file():
        with gitignore.open(encoding="utf-8") as gf:
            lines = gf.readlines()
    return PathSpec.from_lines("gitwildmatch", lines)


def get_files(config):
    gitignore = get_gitignore(config.root)
    for file in config.paths:
        yield from get_absolute_path(Path(file), config, gitignore)


def get_absolute_path(path, config, gitignore):
    if not path.exists():
        raise FileError(path)
    if config.is_path_ignored(path):
        return
    if gitignore is not None and gitignore.match_file(path):
        return
    if path.is_file():
        if should_parse(config, path):
            yield path.absolute()
    elif path.is_dir():
        for file in path.iterdir():
            if file.is_dir() and not config.recursive:
                continue
            yield from get_absolute_path(
                file,
                config,
                gitignore + get_gitignore(path) if gitignore is not None else None,
            )


def should_parse(config, file):
    """Check if file extension is in list of supported file types (can be configured from cli)"""
    return file.suffix and file.suffix.lower() in config.filetypes



"""
Transformers are classes used to transform passed Robot Framework code model.

To create your own transformer you need to create file with the same name as your transformer class. Your class
need to inherit from ``ModelTransformer`` or ``ast.NodeTransformer`` class. Finally put name of your transformer in
``TRANSFORMERS`` variable in this file.

If you don't want to run your transformer by default and only when calling robotidy with --transform YourTransformer
then add ``ENABLED = False`` class attribute inside.
"""
from itertools import chain
from robot.utils.importer import Importer
from robot.errors import DataError

from robotidy.utils import RecommendationFinder
from robotidy.exceptions import InvalidParameterError, InvalidParameterFormatError, ImportTransformerError


TRANSFORMERS = [
    "AddMissingEnd",
    "NormalizeSeparators",
    "DiscardEmptySections",
    "MergeAndOrderSections",
    "RemoveEmptySettings",
    "NormalizeAssignments",
    "OrderSettings",
    "OrderSettingsSection",
    "NormalizeTags",
    "OrderTags",
    "AlignSettingsSection",
    "AlignVariablesSection",
    "AlignTestCases",
    "NormalizeNewLines",
    "NormalizeSectionHeaderName",
    "NormalizeSettingName",
    "ReplaceRunKeywordIf",
    "SplitTooLongLine",
    "SmartSortKeywords",
    "RenameTestCases",
    "RenameKeywords",
    "ReplaceReturns",
    "ReplaceBreakContinue",
    "InlineIf",
]


IMPORTER = Importer()


def import_transformer(name, args):
    short_name = name.split(".")[-1]
    try:
        imported_class = IMPORTER.import_class_or_module(name)
        spec = IMPORTER._get_arg_spec(imported_class)
        positional, named = resolve_args(short_name, spec, args)
    except DataError:
        similar_finder = RecommendationFinder()
        similar = similar_finder.find_similar(short_name, TRANSFORMERS)
        raise ImportTransformerError(
            f"Importing transformer '{short_name}' failed. "
            f"Verify if correct name or configuration was provided.{similar}"
        ) from None
    return imported_class(*positional, **dict(named))


def resolve_args(transformer, spec, args):
    if args and not spec.argument_names:
        raise InvalidParameterError(transformer, " This transformer does not accept arguments but they were provided.")
    arg_names = [arg.split("=")[0] for arg in args]
    for arg in arg_names:
        # it's fine to only check for first non-matching parameter
        if arg not in spec.argument_names:
            similar_finder = RecommendationFinder()
            similar = similar_finder.find_similar(arg, spec.argument_names)
            if not similar:
                arg_names = "\n".join(spec.argument_names)
                similar = f" This transformer accepts following arguments: {arg_names}"
            raise InvalidParameterError(transformer, similar) from None
    try:
        return spec.resolve(args)
    except ValueError as err:
        raise InvalidParameterError(transformer, f" {err}") from None


def load_transformer(name, args):
    if not args.get("enabled", True):
        return None
    args = [f"{key}={value}" for key, value in args.items() if key != "enabled"]
    import_name = f"robotidy.transformers.{name}" if name in TRANSFORMERS else name
    return import_transformer(import_name, args)


def join_configs(args, config):
    # args are from --transform Name:param=value and config is from --configure
    temp_args = {}
    for arg in chain(args, config):
        param, value = arg.split("=", maxsplit=1)
        if param == "enabled":
            temp_args[param] = value.lower() == "true"
        else:
            temp_args[param] = value
    return temp_args


def get_args(transformer, allowed_mapped, config):
    try:
        return join_configs(allowed_mapped.get(transformer, ()), config.get(transformer, ()))
    except ValueError:
        raise InvalidParameterFormatError(transformer) from None


def validate_config(config, allowed_mapped):
    for transformer in config:
        if transformer in allowed_mapped or transformer in TRANSFORMERS:
            continue
        similar_finder = RecommendationFinder()
        similar = similar_finder.find_similar(transformer, TRANSFORMERS + list(allowed_mapped.keys()))
        raise ImportTransformerError(
            f"Configuring transformer '{transformer}' failed. " f"Verify if correct name was provided.{similar}"
        ) from None


def load_transformers(allowed_transformers, config, allow_disabled=False, force_order=False):
    """Dynamically load all classes from this file with attribute `name` defined in allowed_transformers"""
    loaded_transformers = []
    allowed_mapped = {name: args for name, args in allowed_transformers} if allowed_transformers else {}
    validate_config(config, allowed_mapped)
    if not force_order:
        for name in TRANSFORMERS:
            if not allowed_mapped or name in allowed_mapped:
                args = get_args(name, allowed_mapped, config)
                imported_class = load_transformer(name, args)
                if imported_class is None:
                    continue
                enabled = getattr(imported_class, "ENABLED", True) or args.get("enabled", False)
                if allowed_mapped or allow_disabled or enabled:
                    loaded_transformers.append(imported_class)
    for name in allowed_mapped:
        if force_order or name not in TRANSFORMERS:
            args = get_args(name, allowed_mapped, config)
            imported_class = load_transformer(name, args)
            if imported_class is not None:
                loaded_transformers.append(imported_class)
    return loaded_transformers



# coding: utf-8
import os
import sys

import pytest

from _pydev_bundle.pydev_monkey import pydev_src_dir
from _pydevd_bundle.pydevd_constants import sorted_dict_repr
from pydevd import SetupHolder

try:
    from _pydev_bundle import pydev_monkey
except:
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))
    from _pydev_bundle import pydev_monkey


@pytest.fixture(autouse=True)
def save_setup_holder():
    original = SetupHolder.setup
    yield
    SetupHolder.setup = original


def test_monkey():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True}
    check = '''C:\\bin\\python.exe -u -c connect(\\"127.0.0.1\\")'''
    debug_command = (
        'import sys; '
        'sys.path.insert(0, r\'%s\'); '
        "import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        "pydevd.settrace(host='127.0.0.1', port=0, suspend=False, "
        'trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command

    assert 'C:\\bin\\python.exe -u -c %s' % debug_command == pydev_monkey.patch_arg_str_win(check)


def test_str_to_args_windows():
    assert ['a', 'b'] == pydev_monkey.str_to_args_windows('a "b"')


def test_monkey_patch_return_original_args():
    check = ['echo', '"my"', '"args"']
    res = pydev_monkey.patch_args(check[:])
    assert res == check


def test_monkey_patch_pathlib_args():
    try:
        import pathlib
    except ImportError:
        pytest.skip('pathlib not available.')

    check = [pathlib.Path('echo'), '"my"', '"args"']
    res = pydev_monkey.patch_args(check[:])
    assert res == check


def test_monkey_patch_wrong_object_type():
    check = [1, 22, '"my"', '"args"']
    res = pydev_monkey.patch_args(check[:])
    assert res == check


def test_monkey_patch_wrong_object_type_2():
    check = ['C:\\bin\\python.exe', '-u', 1, '-qcconnect("127.0.0.1")']
    res = pydev_monkey.patch_args(check[:])
    assert res == check


def test_monkey_patch_args_module_subprocess_pathlib():
    try:
        import pathlib
    except ImportError:
        pytest.skip('pathlib not available.')

    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'multiprocess': True}
    if sys.platform == 'win32':
        python_path = 'C:\\bin\\python.exe'
    else:
        python_path = '/bin/python'
    check = [pathlib.Path(python_path), '-mtest', pathlib.Path('bar')]
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file
    assert pydev_monkey.patch_args(check) == [
        python_path,
        get_pydevd_file(),
        '--module',
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--multiprocess',
        '--protocol-quoted-line',
        '--file',
        'test',
        'bar',
    ]


def test_monkey_patch_args_indc():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-u', '-c', 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-u',
        '-c',
        debug_command
    ]


def test_separate_future_imports():
    found = pydev_monkey._separate_future_imports('''from __future__ import print_function\nprint(1)''')
    assert found == ('from __future__ import print_function;', '\nprint(1)')

    found = pydev_monkey._separate_future_imports('''from __future__ import print_function;print(1)''')
    assert found == ('from __future__ import print_function;', 'print(1)')

    found = pydev_monkey._separate_future_imports('''from __future__ import (\nprint_function);print(1)''')
    assert found == ('from __future__ import (\nprint_function);', 'print(1)')

    found = pydev_monkey._separate_future_imports('''"line";from __future__ import (\n\nprint_function, absolute_imports\n);print(1)''')
    assert found == ('"line";from __future__ import (\n\nprint_function, absolute_imports\n);', 'print(1)')

    found = pydev_monkey._separate_future_imports('''from __future__ import bar\nfrom __future__ import (\n\nprint_function, absolute_imports\n);print(1)''')
    assert found == ('from __future__ import bar\nfrom __future__ import (\n\nprint_function, absolute_imports\n);', 'print(1)')


def test_monkey_patch_args_indc_future_import():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-u', '-c', 'from __future__ import print_function;connect("127.0.0.1")']
    debug_command = (
        "from __future__ import print_function;import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-u',
        '-c',
        debug_command
    ]


def test_monkey_patch_args_indc_future_import2():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-u', '-c', 'from __future__ import print_function\nconnect("127.0.0.1")']
    debug_command = (
        "from __future__ import print_function;import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        '\nconnect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-u',
        '-c',
        debug_command
    ]


def test_monkey_patch_args_indc2():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-u', '-qcconnect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-u',
        '-qc',
        debug_command
    ]


def test_monkey_patch_args_x_flag():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-X', 'faulthandler', '-c', 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-X',
        'faulthandler',
        '-c',
        debug_command
    ]


def test_monkey_patch_args_flag_in_single_arg_1():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-qX', 'faulthandler', '-c', 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-qX',
        'faulthandler',
        '-c',
        debug_command
    ]


def test_monkey_patch_args_flag_in_single_arg_2():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-qX', 'faulthandler', '-c', 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-qX',
        'faulthandler',
        '-c',
        debug_command
    ]


def test_monkey_patch_args_flag_in_single_arg_3():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-qc', 'connect("127.0.0.1")']
    debug_command = (
        "import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-qc',
        debug_command
    ]


def test_monkey_patch_args_x_flag_inline():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-Xfaulthandler', '-c', 'connect("127.0.0.1")', 'arg1']
    debug_command = (
        "import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-Xfaulthandler',
        '-c',
        debug_command,
        'arg1'
    ]


def test_monkey_patch_args_c_flag_inline():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'ppid': os.getpid(), 'protocol-quoted-line': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-X', 'faulthandler', '-cconnect("127.0.0.1")', 'arg1']
    debug_command = (
        "import sys; sys.path.insert(0, r\'%s\'); import pydevd; pydevd.PydevdCustomization.DEFAULT_PROTOCOL='quoted-line'; "
        'pydevd.settrace(host=\'127.0.0.1\', port=0, suspend=False, trace_only_current_thread=False, patch_multiprocessing=True, access_token=None, client_access_token=None, __setup_holder__=%s); '
        ''
        'connect("127.0.0.1")') % (pydev_src_dir, sorted_dict_repr(SetupHolder.setup))
    if sys.platform == "win32":
        debug_command = debug_command.replace('"', '\\"')
        debug_command = '"%s"' % debug_command
    res = pydev_monkey.patch_args(check)
    assert res == [
        'C:\\bin\\python.exe',
        '-X',
        'faulthandler',
        '-c',
        debug_command,
        'arg1'
    ]


def test_monkey_patch_args_module():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'multiprocess': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-m', 'test']
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file
    assert pydev_monkey.patch_args(check) == [
        'C:\\bin\\python.exe',
        get_pydevd_file(),
        '--module',
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--multiprocess',
        '--skip-notify-stdin',
        '--protocol-quoted-line',
        '--file',
        'test',
    ]


def test_monkey_patch_args_unbuffered_module():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'multiprocess': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-u', '-m', 'test']
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file
    assert pydev_monkey.patch_args(check) == [
        'C:\\bin\\python.exe',
        '-u',
        get_pydevd_file(),
        '--module',
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--multiprocess',
        '--skip-notify-stdin',
        '--protocol-quoted-line',
        '--file',
        'test',
    ]


def test_monkey_patch_args_module_inline():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'multiprocess': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-qOmtest']
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file
    assert pydev_monkey.patch_args(check) == [
        'C:\\bin\\python.exe',
        '-qO',
        get_pydevd_file(),
        '--module',
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--multiprocess',
        '--skip-notify-stdin',
        '--protocol-quoted-line',
        '--file',
        'test',
    ]


def test_monkey_patch_args_module_inline2():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'multiprocess': True, 'skip-notify-stdin': True}
    check = ['C:\\bin\\python.exe', '-qOm', 'test']
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file
    assert pydev_monkey.patch_args(check) == [
        'C:\\bin\\python.exe',
        '-qO',
        get_pydevd_file(),
        '--module',
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--multiprocess',
        '--skip-notify-stdin',
        '--protocol-quoted-line',
        '--file',
        'test',
    ]


def test_monkey_patch_args_no_indc():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0'}
    check = ['C:\\bin\\python.exe', 'connect(\\"127.0.0.1\\")', 'with spaces']
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file
    assert pydev_monkey.patch_args(check) == [
        'C:\\bin\\python.exe',
        get_pydevd_file(),
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--protocol-quoted-line',
        '--file',
        '"connect(\\\\\\"127.0.0.1\\\\\\")"' if sys.platform == 'win32' else 'connect(\\"127.0.0.1\\")',
        '"with spaces"'  if sys.platform == 'win32' else 'with spaces',
    ]


def test_monkey_patch_args_no_indc_with_pydevd():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0'}
    check = ['C:\\bin\\python.exe', 'pydevd.py', 'connect(\\"127.0.0.1\\")', 'bar']

    assert pydev_monkey.patch_args(check) == [
        'C:\\bin\\python.exe', 'pydevd.py', 'connect(\\"127.0.0.1\\")', 'bar']


def test_monkey_patch_args_no_indc_without_pydevd():
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0'}
    check = ['C:\\bin\\python.exe', 'target.py', 'connect(\\"127.0.0.1\\")', 'bar']
    assert pydev_monkey.patch_args(check) == [
        'C:\\bin\\python.exe',
        get_pydevd_file(),
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--protocol-quoted-line',
        '--file',
        'target.py',
        '"connect(\\\\\\"127.0.0.1\\\\\\")"' if sys.platform == 'win32' else 'connect(\\"127.0.0.1\\")',
        'bar',
    ]


@pytest.mark.parametrize('use_bytes', [True, False])
def test_monkey_patch_c_program_arg(use_bytes):
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file

    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'module': 'ignore.this'}
    check = ['C:\\bin\\python.exe', '-u', 'target.py', '-c', '-áéíóú']

    encode = lambda s:s
    if use_bytes:
        check = [c.encode('utf-8') for c in check]
        encode = lambda s:s.encode('utf-8')

    assert pydev_monkey.patch_args(check) == [
        encode('C:\\bin\\python.exe'),
        encode('-u'),
        get_pydevd_file(),
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--protocol-quoted-line',
        '--file',
        encode('target.py'),
        encode('-c'),
        encode('-áéíóú')
    ]


def test_monkey_patch_args_module_single_arg():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'multiprocess': True, 'module': 'ignore.this'}
    check = ['C:\\bin\\python.exe', '-mtest', 'bar']
    from _pydevd_bundle.pydevd_command_line_handling import get_pydevd_file
    assert pydev_monkey.patch_args(check) == [
        'C:\\bin\\python.exe',
        get_pydevd_file(),
        '--module',
        '--port',
        '0',
        '--ppid',
        str(os.getpid()),
        '--client',
        '127.0.0.1',
        '--multiprocess',
        '--protocol-quoted-line',
        '--file',
        'test',
        'bar',
    ]


def test_monkey_patch_args_stdin():
    SetupHolder.setup = {'client': '127.0.0.1', 'port': '0', 'multiprocess': True, 'module': 'ignore.this'}
    check = ['C:\\bin\\python.exe', '-Xfaulthandler' , '-']
    # i.e.: we don't deal with the stdin.
    assert pydev_monkey.patch_args(check) == check



"""
Comments checkers
"""
from codecs import BOM_UTF8, BOM_UTF16_BE, BOM_UTF16_LE, BOM_UTF32_BE, BOM_UTF32_LE

from robot.utils import FileReader

from robocop.checkers import RawFileChecker, VisitorChecker
from robocop.rules import Rule, RuleSeverity
from robocop.utils import ROBOT_VERSION

rules = {
    "0701": Rule(
        rule_id="0701",
        name="todo-in-comment",
        msg="Found {{ todo_or_fixme }} in comment",
        severity=RuleSeverity.WARNING,
        docs="""
        Example::
        
            # TODO: Refactor this code
            # fixme 
        
        """,
    ),
    "0702": Rule(
        rule_id="0702",
        name="missing-space-after-comment",
        msg="Missing blank space after comment character",
        severity=RuleSeverity.WARNING,
        docs="""
        Make sure to have one blank space after '#' comment character

        Example::
        
            #bad
            # good
        """,
    ),
    "0703": Rule(
        rule_id="0703",
        name="invalid-comment",
        msg="Invalid comment. '#' needs to be first character in the cell. "
        "For block comments you can use '*** Comments ***' section",
        severity=RuleSeverity.ERROR,
        version="<4.0",
        docs="""
        In Robot Framework 3.2.2 comments that started from second character in the cell were not recognized as 
        comments.
        
        Example::
        
            # good
             # bad
              # third cell so it's good
        
        """,
    ),
    "0704": Rule(
        rule_id="0704",
        name="ignored-data",
        msg="Ignored data found in file",
        severity=RuleSeverity.WARNING,
        docs="""
        All lines before first test data section 
        (`ref <https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-data-sections>`_) 
        are ignored. It's recommended to add `*** Comments ***` section header for lines that should be ignored.
        
        Missing section header::
    
            Resource   file.resource  # it looks like *** Settings *** but section header is missing - line is ignored
            
            *** Keywords ***
            Keyword Name
               No Operation
        
        Comment lines that should be inside `*** Comments ***`::
            
            Deprecated Test
                Keyword
                Keyword 2
            
            *** Test Cases ***
    
        """,
    ),
    "0705": Rule(
        rule_id="0705",
        name="bom-encoding-in-file",
        msg="This file contains BOM (Byte Order Mark) encoding not supported by Robot Framework",
        severity=RuleSeverity.WARNING,
        docs="""
        Some code editors can save Robot file using BOM encoding. Ensure that file is saved in UTF-8 encoding.
        """,
    ),
}


class CommentChecker(VisitorChecker):
    """Checker for comments content. It detects invalid comments or leftovers like `todo` or `fixme` in the code."""

    reports = (
        "todo-in-comment",
        "missing-space-after-comment",
        "invalid-comment",
    )

    def visit_Comment(self, node):  # noqa
        self.find_comments(node)

    def visit_TestCase(self, node):  # noqa
        self.check_invalid_comments(node.name, node)
        self.generic_visit(node)

    visit_Keyword = visit_TestCase

    def visit_Statement(self, node):  # noqa
        self.find_comments(node)

    def find_comments(self, node):
        for token in node.tokens:
            if token.type == "COMMENT":
                self.check_comment_content(token)

    def check_invalid_comments(self, name, node):
        if ROBOT_VERSION.major != 3:
            return
        if name and name.lstrip().startswith("#"):
            hash_pos = name.find("#")
            self.report("invalid-comment", node=node, col=node.col_offset + hash_pos + 1)

    def check_comment_content(self, token):
        if "todo" in token.value.lower():
            self.report(
                "todo-in-comment",
                todo_or_fixme="TODO",
                lineno=token.lineno,
                col=token.col_offset + 1 + token.value.lower().find("todo"),
            )
        if "fixme" in token.value.lower():
            self.report(
                "todo-in-comment",
                todo_or_fixme="FIXME",
                lineno=token.lineno,
                col=token.col_offset + 1 + token.value.lower().find("fixme"),
            )
        if token.value.startswith("#") and token.value != "#":
            if not token.value.startswith("# "):
                self.report(
                    "missing-space-after-comment",
                    lineno=token.lineno,
                    col=token.col_offset + 1,
                )


class IgnoredDataChecker(RawFileChecker):
    """Checker for ignored data."""

    reports = (
        "ignored-data",
        "bom-encoding-in-file",
    )
    BOM = [BOM_UTF32_BE, BOM_UTF32_LE, BOM_UTF8, BOM_UTF16_LE, BOM_UTF16_BE]

    def __init__(self):
        self.is_bom = False
        super().__init__()

    def parse_file(self):
        self.is_bom = False
        if self.lines is not None:
            for lineno, line in enumerate(self.lines, start=1):
                if self.check_line(line, lineno):
                    break
        else:
            self.detect_bom(self.source)
            with FileReader(self.source) as file_reader:
                for lineno, line in enumerate(file_reader.readlines(), start=1):
                    if self.check_line(line, lineno):
                        break

    def check_line(self, line, lineno):
        if line.startswith("***"):
            return True
        if not line.startswith("# robocop:"):
            if lineno == 1 and self.is_bom:
                # if it's BOM encoded file, first line can be ignored
                return "***" in line
            self.report("ignored-data", lineno=lineno, col=1)
            return True
        return False

    def detect_bom(self, source):
        with open(source, "rb") as raw_file:
            first_four = raw_file.read(4)
            self.is_bom = any(first_four.startswith(bom_marker) for bom_marker in IgnoredDataChecker.BOM)
            if self.is_bom:
                self.report("bom-encoding-in-file", lineno=1, col=1)



def ask_for_stop(use_back):
    import pydevd
    if use_back:
        pydevd.settrace(stop_at_frame=sys._getframe().f_back)
    else:
        pydevd.settrace()
    print('Will stop here if use_back==False.')


def outer_method():
    ask_for_stop(True)
    print('will stop here.')
    ask_for_stop(False)


if __name__ == '__main__':
    import os
    import sys
    root_dirname = os.path.dirname(os.path.dirname(__file__))
    
    if root_dirname not in sys.path:
        sys.path.append(root_dirname)

    outer_method()        
    print('TEST SUCEEDED!')
    
